# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.


from typing import Iterable, TypeVar

import torch

from torcheval.metrics.metric import Metric

TCat = TypeVar("TCat")


class Cat(Metric[torch.Tensor]):
    """
    Concatenate all input tensors along dimension dim. Its functional
    version is ``torch.cat(input)``.

    All input tensors to ``Cat.update()`` must either have the same shape
    (except in the concatenating dimension) or be empty.

    Zero-dimensional tensor is not a valid input of ``Cat.update()``.
    ``torch.flatten()`` can be used to flatten zero-dimensional into
    an one-dimensional tensor before passing in ``Cat.update()``.

    Example:
        >>> import torch
        >>> from torcheval.metrics import Cat
        >>> metric = Cat(dim=1)
        >>> metric.update(torch.tensor([[1, 2], [3, 4]]))
        >>> metric.compute()
        tensor([[1, 2],
                [3, 4]]))

        >>> metric.update(torch.tensor([[5, 6], [7, 8]]))).compute()
        tensor([[1, 2, 5, 6],
                [3, 4, 7, 8]]))

        >>> metric.reset()
        >>> metric.update(torch.tensor([0])).compute()
        tensor([0])
    """

    def __init__(self: "Cat", dim: int = 0) -> None:
        """
        Initialize a Cat metric object.

        Args:
            dim: The dimension along which to concatenate, as in ``torch.cat()``.
        """
        super().__init__()
        self.dim = dim
        self._add_state("inputs", [])

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(self: TCat, input: torch.Tensor) -> TCat:
        self.inputs.append(input)
        return self

    @torch.inference_mode()
    def compute(self: TCat) -> torch.Tensor:
        """
        Return the concatenated inputs.

        If no calls to ``update()`` are made before ``compute()`` is called,
        the function returns ``torch.empty(0)``.
        """
        if not self.inputs:
            return torch.empty(0)
        return torch.cat(self.inputs, dim=self.dim)

    @torch.inference_mode()
    def merge_state(self: TCat, metrics: Iterable[TCat]) -> TCat:
        for metric in metrics:
            if metric.inputs:
                self.inputs.append(
                    torch.cat(metric.inputs, dim=metric.dim).to(self.device)
                )
        return self

    @torch.inference_mode()
    def _prepare_for_merge_state(self: TCat) -> None:
        if self.inputs:
            self.inputs = [torch.cat(self.inputs, dim=self.dim)]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, Optional, TypeVar

import torch

from torcheval.metrics.functional import reciprocal_rank
from torcheval.metrics.metric import Metric


TReciprocalRank = TypeVar("TReciprocalRank")


class ReciprocalRank(Metric[torch.Tensor]):
    """
    Compute the reciprocal rank of the correct class among the top predicted classes.
    Its functional version is ``torcheval.metrics.functional.reciprocal_rank``.

    Args:
        k (int, optional): Number of top class probabilities to be considered.

    Example:
        >>> import torch
        >>> from torcheval.metrics import ReciprocalRank

        >>> metric = ReciprocalRank()
        >>> metric.update(torch.tensor([[0.3, 0.1, 0.6], [0.5, 0.2, 0.3]]), torch.tensor([2, 1]))
        >>> metric.update(torch.tensor([[0.2, 0.1, 0.7], [0.3, 0.3, 0.4]]), torch.tensor([1, 0]))
        >>> metric.compute()
        tensor([1.0000, 0.3333, 0.3333, 0.5000])

        >>> metric = ReciprocalRank(k=2)
        >>> metric.update(torch.tensor([[0.3, 0.1, 0.6], [0.5, 0.2, 0.3]]), torch.tensor([2, 1]))
        >>> metric.update(torch.tensor([[0.2, 0.1, 0.7], [0.3, 0.3, 0.4]]), torch.tensor([1, 0]))
        >>> metric.compute()
        tensor([1.0000, 0.0000, 0.0000, 0.5000])
    """

    def __init__(self: TReciprocalRank, k: Optional[int] = None) -> None:
        super().__init__()
        self.k = k
        self._add_state("scores", [])

    @torch.inference_mode()
    # pyre-ignore[14]: `update` overrides method defined in `Metric` inconsistently.
    def update(
        self: TReciprocalRank, input: torch.Tensor, target: torch.Tensor
    ) -> TReciprocalRank:
        """
        Update the metric state with the ground truth labels and predictions.

        Args:
            input (Tensor): Predicted unnormalized scores (often referred to as logits) or
                class probabilities of shape (num_samples, num_classes).
            target (Tensor): Ground truth class indices of shape (num_samples,).
        """
        self.scores.append(reciprocal_rank(input, target, k=self.k))
        return self

    @torch.inference_mode()
    def compute(self: TReciprocalRank) -> torch.Tensor:
        """
        Return the concatenated reciprocal rank scores. If no ``update()`` calls are made before
        ``compute()`` is called, return an empty tensor.
        """
        if not self.scores:
            return torch.empty(0)
        return torch.cat(self.scores, dim=0)

    @torch.inference_mode()
    def merge_state(
        self: TReciprocalRank, metrics: Iterable[TReciprocalRank]
    ) -> TReciprocalRank:
        """
        Merge the metric state with its counterparts from other metric instances.

        Args:
            metrics (Iterable[Metric]): metric instances whose states are to be merged.
        """
        for metric in metrics:
            if metric.scores:
                self.scores.append(torch.cat(metric.scores).to(self.device))
        return self

    @torch.inference_mode()
    def _prepare_for_merge_state(self: TReciprocalRank) -> None:
        if self.scores:
            self.scores = [torch.cat(self.scores)]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from torcheval.metrics.functional.classification.accuracy import (
    binary_accuracy,
    multiclass_accuracy,
    multilabel_accuracy,
)
from torcheval.metrics.functional.classification.auroc import binary_auroc
from torcheval.metrics.functional.classification.binned_precision_recall_curve import (
    binary_binned_precision_recall_curve,
)
from torcheval.metrics.functional.classification.f1_score import multiclass_f1_score
from torcheval.metrics.functional.classification.precision import multiclass_precision
from torcheval.metrics.functional.classification.precision_recall_curve import (
    binary_precision_recall_curve,
    multiclass_precision_recall_curve,
)
from torcheval.metrics.functional.classification.recall import multiclass_recall

__all__ = [
    "binary_auroc",
    "binary_accuracy",
    "binary_precision_recall_curve",
    "binary_binned_precision_recall_curve",
    "multiclass_accuracy",
    "multiclass_f1_score",
    "multiclass_precision",
    "multiclass_precision_recall_curve",
    "multiclass_recall",
    "multilabel_accuracy",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import copy
import math
import warnings
from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional, Tuple

import torch
from torch.nn.parameter import UninitializedParameter
from torcheval.tools.flops import (
    flop_mapping,
    FlopTensor,
    instrument_module,
    start_counting,
)

_ATTRIB_TO_COL_HEADER = {
    "module_name": "Name",
    "module_type": "Type",
    "num_parameters": "# Params",
    "num_trainable_parameters": "# Trainable Params",
    "has_uninitialized_param": "Contains Uninitialized Param?",
    "flops_forward": "Forward FLOPs",
    "flops_backward": "Backward FLOPs",
}  # Attribute: column header (in table)
_ATTRIBS: List[str] = list(_ATTRIB_TO_COL_HEADER.keys())
_FLOP_ATTRIBS: List[str] = ["flops_forward", "flops_backward"]


_PARAMETER_NUM_UNITS = [" ", "K", "M", "B", "T"]
_PARAMETER_FLOPS_UNITS = [" ", "k", "M", "G", "T", "P", "E", "Z", "Y"]


class ModuleSummary:
    """
    Summary of module and its submodules. It collects the following information:
    - Name
    - Type
    - Number of parameters
    - Number of trainable parameters
    - FLOPs for forward (-1 meaning not calculated)
    - FLOPs for backward (-1 meaning not calculated)
    """

    def __init__(self) -> None:
        self._module_name: str = ""
        self._module_type: str = ""
        self._num_parameters: int = 0
        self._num_trainable_parameters: int = 0
        self._submodule_summaries: Dict[str, "ModuleSummary"] = {}
        self._has_uninitialized_param: bool = False
        self._flops_forward: int = -1
        self._flops_backward: int = -1
        self._flops_forward_detail: Dict[str, int] = {}
        self._flops_backward_detail: Dict[str, int] = {}

    @property
    def submodule_summaries(self) -> Dict[str, "ModuleSummary"]:
        """
        A Dict with the names of submodules as keys and corresponding ModuleSummary
        objects as values. These can be traversed for visualization.
        """
        return self._submodule_summaries

    @property
    def module_name(self) -> str:
        """Returns the name of this module"""
        return self._module_name

    @property
    def module_type(self) -> str:
        """Returns the type of this module."""
        return self._module_type

    @property
    def num_parameters(self) -> int:
        """Returns the total number of parameters in this module."""
        if self.has_uninitialized_param:
            warnings.warn(
                "A layer with UninitializedParameter was found. "
                "Thus, the total number of parameters detected may be inaccurate."
            )
        return self._num_parameters

    @property
    def num_trainable_parameters(self) -> int:
        """
        Returns the total number of trainable parameters (requires_grad=True)
        in this module.
        """
        if self.has_uninitialized_param:
            warnings.warn(
                "A layer with UninitializedParameter was found. "
                "Thus, the total number of parameters detected may be inaccurate."
            )
        return self._num_trainable_parameters

    @property
    def flops_forward(self) -> int:
        """Returns the total FLOPs for forward calculation using this module."""
        if self.has_uninitialized_param:
            warnings.warn(
                "A layer with UninitializedParameter was found. "
                "Thus, the total number of FLOPs detected may be inaccurate."
            )
        return self._flops_forward

    @property
    def flops_backward(self) -> int:
        """Returns the total FLOPs for backward calculation using this module."""
        if self.has_uninitialized_param:
            warnings.warn(
                "A layer with UninitializedParameter was found. "
                "Thus, the total number of FLOPs detected may be inaccurate."
            )
        return self._flops_backward

    @property
    def has_uninitialized_param(self) -> bool:
        """Returns if a parameter in this module is uninitialized"""
        return self._has_uninitialized_param

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return _get_summary_table(self)


def _clean_flops(flop: DefaultDict[str, DefaultDict[str, int]], N: int) -> None:
    for _, sub_flop in flop.items():
        for opr in sub_flop:
            sub_flop[opr] = sub_flop[opr] // N


def _get_module_flops(
    module: torch.nn.Module, module_input: torch.Tensor
) -> Tuple[
    DefaultDict[str, DefaultDict[str, int]], DefaultDict[str, DefaultDict[str, int]]
]:
    # TODO: here we assume module_input should be a single tensor and module's output should be a single tensor.
    #   Need to provide more flexible implementation that support other type of input/output.
    all_hooks: List[torch.utils.hooks.RemovableHandle] = []
    instrument_module(module, all_hooks, "")
    module_input = FlopTensor(module_input)
    N = len(module_input)
    module.zero_grad()

    # Count for forward flops
    start_counting()
    res = module(module_input).mean()
    flops_forward = copy.deepcopy(FlopTensor.flop_counts)
    # Count for backward flops
    start_counting()
    res.backward()
    flops_backward = copy.deepcopy(FlopTensor.flop_counts)

    # Norm FLOPs if N>1
    if N > 1:
        _clean_flops(flops_forward, N)
        _clean_flops(flops_backward, N)

    # Reverting all the changes:
    module.zero_grad()
    # TODO: Reverting BN: We also need to save status of BN running mean/var before running and revert those.
    for hood_handle in all_hooks:
        hood_handle.remove()
    return flops_forward, flops_backward


def _has_uninitialized_param(module: torch.nn.Module) -> bool:
    for param in module.parameters():
        if isinstance(param, UninitializedParameter):
            return True
    return False


def get_module_summary(
    module: torch.nn.Module,
    max_depth: Optional[int] = None,
    module_input: Optional[torch.Tensor] = None,
) -> ModuleSummary:
    """
    Generate a ModuleSummary object, then assign its values and generate submodule tree.

    Args:
        module: The module to be summarized.
        max_depth: The maximum depth of modules to generate tree.
        Value must be greater than 0. Set to `None` to include
        all submodules.
        module_name: The name of the current module.
        module_input: An input for the module to run and calculate FLOPs.
        Note: if module contains any lazy submodule, we will NOT calculate FLOPs.
        It should only contain 1 sample (i.e. batch_size=1), otherwise, we will return FLOPs = total FLOPs / len(module_input).
        Note: currently we only support module that takes a single tensor (i.e. module_input should be a torch.Tensor)
            and outputs a single tensor. TODO: to support more flexible input and output for module.

    Raises:
        ValueError:
            If `max_depth` is an int less than 1
    """
    if max_depth is not None and max_depth < 1:
        raise ValueError(
            f"`max_depth` must be `None` or an int greater than 0. Got {max_depth}."
        )
    flops_forward = None
    flops_backward = None
    has_uninitialized_param = _has_uninitialized_param(module)
    if module_input is not None and not has_uninitialized_param:
        flops_forward, flops_backward = _get_module_flops(module, module_input)
    return _generate_module_summary(
        module,
        max_depth,
        "",
        flops_forward=flops_forward,
        flops_backward=flops_backward,
    )


def _generate_module_summary(
    module: torch.nn.Module,
    max_depth: Optional[int],
    module_name: str,
    flops_forward: Optional[DefaultDict[str, DefaultDict[str, int]]] = None,
    flops_backward: Optional[DefaultDict[str, DefaultDict[str, int]]] = None,
) -> ModuleSummary:
    """
    Recursively generate and populate metrics for ModelSummary.
    """
    module_summary = ModuleSummary()

    # Populate Submodule Summaries
    submodule_summaries = {}
    new_max_depth = max_depth - 1 if max_depth else None
    if new_max_depth is None or new_max_depth > 0:
        for name, submodule in module.named_children():
            formatted_name = name
            if module_name != "":
                formatted_name = f"{module_name}.{formatted_name}"
            submodule_summaries[formatted_name] = _generate_module_summary(
                submodule,
                new_max_depth,
                formatted_name,
                flops_forward=flops_forward,
                flops_backward=flops_backward,
            )
    # Populate values in module_summary
    module_summary._submodule_summaries = submodule_summaries
    module_summary._module_name = module_name
    module_summary._module_type = str(module.__class__.__name__)
    for param in module.parameters():
        if isinstance(param, UninitializedParameter):
            module_summary._has_uninitialized_param = True
        else:
            numel = param.numel()
            module_summary._num_parameters += numel
            if param.requires_grad:
                module_summary._num_trainable_parameters += numel
    # Calculate flops forward/backward.
    if flops_forward is not None:
        module_summary._flops_forward_detail = dict(flops_forward[module_name])
        module_summary._flops_forward = sum(
            [v for k, v in flops_forward[module_name].items()]
        )
    if flops_backward is not None:
        module_summary._flops_backward_detail = dict(flops_backward[module_name])
        module_summary._flops_backward = sum(
            [v for k, v in flops_backward[module_name].items()]
        )

    return module_summary


def _get_summary_table(
    module_summary: ModuleSummary, human_readable_nums: bool = True
) -> str:
    """
    Generates a string summary_table, tabularizing the information in module_summary.

    Args:
        module_summary: module_summary to be printed/tabularized
        human_readable_nums: set to False for exact (e.g. 1234 vs 1.2 K)
    """
    stop_attr: List[str] = []
    # Unpack attributes
    if module_summary.flops_forward == -1:
        stop_attr.append("flops_forward")
    if module_summary.flops_backward == -1:
        stop_attr.append("flops_backward")
    unpacked_attribs, col_widths = defaultdict(list), defaultdict(int)
    _unpack_attributes(
        {"root": module_summary},
        unpacked_attribs,
        col_widths,
        human_readable_nums,
        stop_attr,
    )

    # Generate formatted summary_table string
    s = "{:{}}"  # inner {}: col_width
    use_attribs = [attr for attr in _ATTRIBS if attr not in stop_attr]
    n_rows, n_cols = len(unpacked_attribs[use_attribs[0]]), len(use_attribs)
    total_width = sum(col_widths.values()) + 3 * (n_cols - 1)

    header = [
        s.format(col_header, col_width)
        for col_header, col_width in zip(
            _ATTRIB_TO_COL_HEADER.values(), col_widths.values()
        )
    ]
    summary_table = " | ".join(header) + "\n" + "-" * total_width + "\n"

    for i in range(n_rows):
        row = []
        for attrib in use_attribs:
            row.append(unpacked_attribs[attrib][i])
            row = [
                s.format(r, col_width) for r, col_width in zip(row, col_widths.values())
            ]
        summary_table += " | ".join(row) + "\n"
    # Add disclaims for FLOPs:
    if "flops_forward" not in stop_attr or "flops_backward" not in stop_attr:
        used_operators = "|".join(
            [
                f"`{j.__name__}`"
                for j in flop_mapping.keys()
                if not j.__name__.endswith(".default")
            ]
        )
        summary_table += (
            f"Remark for FLOPs calculation: (1) Only operators {used_operators} are included. "
            + "To add more operators supported in FLOPs calculation, "
            + "please contribute to torcheval/tools/flops.py. "
            + "(2) The calculation related to additional loss function is not included. "
            + "For forward, we calculated FLOPs based on `loss = model(input_data).mean()`. "
            + "For backward, we calculated FLOPs based on `loss.backward()`. \n"
        )
    return summary_table


def _unpack_attributes(
    module_summaries: Dict[str, ModuleSummary],
    unpacked_attribs: Dict[str, List[str]],
    col_widths: Dict[str, int],
    human_readable_nums: bool = True,
    stop_attr: Optional[List[str]] = None,
) -> None:
    """
    Unpack/flatten recursive module_summaries into table columns and store in unpacked_attribs.
    Also, populate col_widths (with max column width).

    Args:
        module_summaries: dict of module summaries
        unpacked_attribs: collects unpacked/flattened columns
        col_widths: tracks max table width for each column
        human_readable_nums: human readable nums (e.g. 1.2 K for 1234)
        stop_attr: a list of attributes that we stop from adding to the table,
        i.e. exclude from _ATTRIBS
    """

    if not module_summaries:
        return

    for module_summary in module_summaries.values():
        for attrib in _ATTRIBS:
            if stop_attr is not None and attrib in stop_attr:
                continue

            # Convert attribute value to string appropriately
            attrib_val = getattr(module_summary, attrib)
            if isinstance(attrib_val, bool):
                attrib_val = "Yes" if attrib_val else "No"
            elif isinstance(attrib_val, int) and attrib in _FLOP_ATTRIBS:
                if attrib_val < 0:
                    attrib_val = ""
                else:
                    attrib_val = (
                        _get_human_readable_count(
                            attrib_val, labels=_PARAMETER_FLOPS_UNITS
                        )
                        if human_readable_nums
                        else str(attrib_val)
                    )
            elif isinstance(attrib_val, int):
                attrib_val = (
                    _get_human_readable_count(attrib_val)
                    if human_readable_nums
                    else str(attrib_val)
                )
            elif attrib_val is None:
                attrib_val = ""

            # Store converted attribute value, track max column width
            unpacked_attribs[attrib].append(attrib_val)
            col_widths[attrib] = max(
                len(_ATTRIB_TO_COL_HEADER[attrib]),
                len(attrib_val),
                col_widths[attrib],
            )
        # Recurse
        _unpack_attributes(
            module_summary.submodule_summaries,
            unpacked_attribs,
            col_widths,
            human_readable_nums,
            stop_attr,
        )


def _get_human_readable_count(number: int, labels: Optional[List[str]] = None) -> str:
    """Abbreviates an integer number with K, M, B, T for thousands, millions, billions and trillions, respectively.
    Examples:
        >>> _get_human_readable_count(123)
        '123  '
        >>> _get_human_readable_count(1234)  # (one thousand)
        '1.2 K'
        >>> _get_human_readable_count(int(2e6))   # (two million)
        '2.0 M'
        >>> _get_human_readable_count(int(3e9))   # (three billion)
        '3.0 B'
        >>> _get_human_readable_count(int(3e9), labels=[" ", "K", "M", "G", "T"])  # (Using units for FLOPs, 3 G)
        '3.0 G'
        >>> _get_human_readable_count(int(4e14))  # (four hundred trillion)
        '400 T'
        >>> _get_human_readable_count(int(5e15))  # (more than trillion)
        '5,000 T'
    Args:
        number: a positive integer number
        labels: a list of units that we want to use per 10^3 digits. Defaults to [" ", "K", "M", "B", "T"]
    Return:
        A string formatted according to the pattern described above.
    Raises:
        ValueError:
            If `number` is less than 0
        TypeError:
            If `number` is not an int
    """
    # logic does not work for floats (e.g. number=0.5)
    if not isinstance(number, int):
        raise TypeError(f"Input type must be int, but received {type(number)}")
    if number < 0:
        raise ValueError(f"Input value must be greater than 0, received {number}")
    labels = labels or _PARAMETER_NUM_UNITS
    if len(labels) <= 0:
        raise ValueError(
            f"Input labels must be a list with at least one string, received {labels}"
        )

    num_digits = int(math.floor(math.log10(number)) + 1 if number > 0 else 1)
    num_groups = int(math.ceil(num_digits / 3))
    num_groups = min(num_groups, len(labels))  # don't abbreviate beyond trillions
    shift = -3 * (num_groups - 1)
    number = number * (10**shift)
    index = num_groups - 1
    if index < 1 or number >= 100:
        return f"{int(number):,d} {labels[index]}"

    return f"{number:,.1f} {labels[index]}"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from torcheval.metrics.classification.accuracy import (
    BinaryAccuracy,
    MulticlassAccuracy,
    MultilabelAccuracy,
)
from torcheval.metrics.classification.auroc import BinaryAUROC
from torcheval.metrics.classification.f1_score import MulticlassF1Score
from torcheval.metrics.classification.precision import MulticlassPrecision
from torcheval.metrics.classification.precision_recall_curve import (
    BinaryPrecisionRecallCurve,
    MulticlassPrecisionRecallCurve,
)
from torcheval.metrics.classification.recall import MulticlassRecall

__all__ = [
    "BinaryAUROC",
    "BinaryAccuracy",
    "BinaryPrecisionRecallCurve",
    "MulticlassAccuracy",
    "MulticlassF1Score",
    "MulticlassPrecision",
    "MulticlassRecall",
    "MultilabelAccuracy",
    "MulticlassPrecisionRecallCurve",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import os
from enum import Enum
from typing import Any, Callable, Dict, Optional

import torch
import torch.distributed as dist
from torcheval.metrics import Metric
from torcheval.metrics.toolkit import sync_and_compute
import torch.utils.benchmark as benchmark


class InputType(Enum):
   # classification
   BINARY_LABEL = "binary_label"
   BINARY_LOGIT = "binary_logit"
   BINARY_PROB = "binary_prob"
   MULTICLASS_LABEL = "multiclass_label"
   MULTICLASS_LOGIT = "multiclass_logit"
   MULTICLASS_PROB = "multiclass_prob"
   MULTILABEL = "multilabel"
   # regression/aggregation
   SINGLE_VALUE = "single_output"
   MULTI_DIM = "multi_dim"


def generate_random_input(
    input_type: InputType,
    device: torch.device,
    num_batches:int,
    batch_size:int,
    num_classes: Optional[int] = None,
    num_dim: Optional[int] = None,
) -> torch.Tensor:
    if input_type == InputType.BINARY_LABEL:
        return torch.randint(high=2, size=(num_batches, batch_size), device=device)
    elif input_type == InputType.BINARY_LOGIT:
        return torch.randn(num_batches, batch_size, device=device)
    elif input_type == InputType.BINARY_PROB:
        return torch.rand(num_batches, batch_size, device=device)
    elif input_type == InputType.MULTICLASS_LABEL:
        return torch.randint(high=num_classes, size=(num_batches, batch_size), device=device)
    elif input_type == InputType.MULTICLASS_LOGIT:
        return torch.randn(num_batches, batch_size, num_classes, device=device)
    elif input_type == InputType.MULTICLASS_PROB:
        return torch.rand(num_batches, batch_size, num_classes, device=device)
    elif input_type == InputType.MULTILABEL:
        return torch.randint(high=num_classes, size=(num_batches, batch_size, num_classes), device=device)
    elif input_type ==  InputType.SINGLE_VALUE:
        return torch.rand(num_batches, batch_size, device=device)
    elif input_type ==  InputType.MULTI_DIM:
        return torch.rand(num_batches, batch_size, num_dim, device=device)


### MAIN BENCHMARK FUNCTION
def get_benchmark_stats(metric: Metric, num_batches:int, compute_interval:int, update_kwargs: Dict[str, Any]):
    metric.reset()
    get_gpu_memory(metric, num_batches, compute_interval, update_kwargs)
    metric.reset()
    get_time(metric, num_batches, compute_interval, update_kwargs)


def get_time(metric: Metric, num_batches:int, compute_interval:int, update_kwargs: Dict[str, Any]):
    timer = benchmark.Timer(
        setup='from __main__ import run_metric_computation',
        stmt='run_metric_computation(metric, num_batches, compute_interval, update_kwargs)',
        globals={
            'metric': metric,
            'num_batches': num_batches,
            'compute_interval': compute_interval,
            'update_kwargs': update_kwargs,
        },
    )
    result = timer.timeit(10)
    rank_0_print(f"Average time spent of 20 runs: {result.mean}")

def get_gpu_memory(metric: Metric, num_batches:int, compute_interval:int, update_kwargs: Dict[str, Any]):
    from __main__ import run_metric_computation

    torch.cuda.empty_cache()
    run_metric_computation(metric, num_batches, compute_interval, update_kwargs)
    rank_0_print("memory usage on rank 0...")
    rank_0_print(torch.cuda.memory_summary(abbreviated=False))


### UTILITIES
def setup_distributed() -> torch.device:
    dist.init_process_group(backend="nccl")
    local_rank = int(os.environ["LOCAL_RANK"])
    device = torch.device(f"cuda:{local_rank}")
    torch.cuda.set_device(device)
    return device

def rank_0_print(msg: str) -> None:
    if dist.get_rank() == 0:
        print(msg)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from torcheval.metrics import functional
from torcheval.metrics.aggregation import Cat, Max, Mean, Min, Sum, Throughput
from torcheval.metrics.classification import (
    BinaryAccuracy,
    BinaryAUROC,
    BinaryPrecisionRecallCurve,
    MulticlassAccuracy,
    MulticlassF1Score,
    MulticlassPrecision,
    MulticlassPrecisionRecallCurve,
    MulticlassRecall,
    MultilabelAccuracy,
)
from torcheval.metrics.metric import Metric
from torcheval.metrics.ranking import ReciprocalRank
from torcheval.metrics.regression import MeanSquaredError, R2Score

__all__ = [
    # base interface
    "Metric",
    # functional metrics
    "functional",
    ## class metrics
    "BinaryAUROC",
    "BinaryAccuracy",
    "BinaryPrecisionRecallCurve",
    "Cat",
    "Max",
    "Mean",
    "MeanSquaredError",
    "Min",
    "MulticlassAccuracy",
    "MulticlassF1Score",
    "MulticlassPrecision",
    "MulticlassPrecisionRecallCurve",
    "MulticlassRecall",
    "MultilabelAccuracy",
    "R2Score",
    "ReciprocalRank",
    "Sum",
    "Throughput",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from torcheval.metrics.aggregation.cat import Cat
from torcheval.metrics.aggregation.max import Max
from torcheval.metrics.aggregation.mean import Mean
from torcheval.metrics.aggregation.min import Min
from torcheval.metrics.aggregation.sum import Sum
from torcheval.metrics.aggregation.throughput import Throughput

__all__ = ["Cat", "Max", "Mean", "Min", "Sum", "Throughput"]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import os
import pickle
import unittest
import uuid
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict, List, Sequence, Set

import torch
import torch.distributed.launcher as pet
from torcheval.metrics import Metric
from torcheval.metrics.toolkit import clone_metric, sync_and_compute
from torcheval.utils.device import copy_data_to_device
from torcheval.utils.env import init_from_env
from typing_extensions import Literal

BATCH_SIZE = 16
# By default, we can test merge_state() on 4 processes with
# each processes will update states twice, which is 8 updates in total.
NUM_TOTAL_UPDATES = 8
NUM_PROCESSES = 4


@dataclass
class _MetricClassTestCaseSpecs:
    metric: Metric
    state_names: Set[str]
    update_kwargs: Dict[str, Any]
    # pyre-ignore[4]: There's no restrictions on return types of a specific metric computation
    compute_result: Any
    # pyre-ignore[4]: There's no restrictions on return types of a specific metric computation
    merge_and_compute_result: Any
    num_total_updates: int = NUM_TOTAL_UPDATES
    num_processes: int = NUM_PROCESSES
    atol: float = 1e-8
    rtol: float = 1e-5
    device: Literal["cuda", "cpu"] = "cpu"


class MetricClassTester(unittest.TestCase):
    def setUp(self) -> None:
        self._test_case_spec = None

    def run_class_implementation_tests(
        self,
        metric: Metric,
        state_names: Set[str],
        update_kwargs: Dict[str, Any],
        # pyre-ignore[2]: There's no restrictions on return types of a specific metric computation
        compute_result: Any,
        # pyre-ignore[2]: There's no restrictions on return types of a specific metric computation
        merge_and_compute_result: Any = None,
        num_total_updates: int = NUM_TOTAL_UPDATES,
        num_processes: int = NUM_PROCESSES,
        atol: float = 1e-8,
        rtol: float = 1e-5,
    ) -> None:
        """
        Run a test case to verify metric class implementations.

        Args:
            metric: The metric object to test against.
            state_names: Set of names of metric state variables.
            update_kwargs: Key value pairs representing the arguments of
                ``metric.update()``. For each argument value, len(val)
                should be equal to ``num_total_updates``, which should be
                greater than 2 to test the behaviour in distributed training.
                Make sure the argument value of ``i``th ``update()`` call
                could be accessed by ``updated_kwargs[arg_name][i]``.
                Usually ``input`` and ``target`` can be generated by
                ``torch.rand(num_updates, BATCH_SIZE, ..) ``.
            compute_result: The expected return value of ``Metric.compute()``
                after metric is updated number of batches times with arguments
                in update_kwargs.
            num_total_updates: Number of updates in the update_kwargs.
            num_processes: Number of processes for metric computation distributed
                training. ``num_total_updates`` should be divisible by
                ``num_processes``.
            atol: Absolute tolerance used in ``torch.testing.assert_close``
            rtol: Relative tolerance used in ``torch.testing.assert_close``
        """
        # update args and state names should not be empty
        self.assertTrue(update_kwargs)
        self.assertTrue(state_names)
        self.assertTrue(
            all(
                len(arg_val) == num_total_updates for arg_val in update_kwargs.values()
            ),
            "The outer size of each update argument should be equal to number of updates",
        )
        self.assertGreater(num_total_updates, 2)
        self.assertGreater(num_processes, 1)
        self.assertEqual(num_total_updates % num_processes, 0)

        merge_and_compute_result = (
            compute_result
            if merge_and_compute_result is None
            else merge_and_compute_result
        )
        self._test_case_spec = _MetricClassTestCaseSpecs(
            metric,
            state_names,
            update_kwargs,
            compute_result,
            merge_and_compute_result,
            num_total_updates,
            num_processes,
            atol,
            rtol,
        )

        test_devices = ("cpu", "cuda") if torch.cuda.is_available() else ("cpu",)
        for device in test_devices:
            self._test_case_spec.device = device
            self._test_case_spec = copy_data_to_device(
                self._test_case_spec, torch.device(self._test_case_spec.device)
            )
            self._test_init()
            self._test_update_and_compute()
            self._test_merge_state()
            # testing on GPU might cause CUDA oom
            if device == "cpu":
                self._test_sync_and_compute()

    def _test_metric_pickable_hashable(self, metric: Metric) -> None:
        pickled_metric = pickle.dumps(metric)
        loaded_metric = pickle.loads(pickled_metric)
        self.assert_state_unchanged(
            self._test_case_spec.state_names, loaded_metric, metric
        )
        self.assertTrue(hash(metric))

    def _test_state_dict_load_state_dict(self, metric: Metric) -> None:
        test_metric = deepcopy(metric).reset()
        test_metric.load_state_dict(metric.state_dict())
        self.assert_state_unchanged(
            self._test_case_spec.state_names, test_metric, metric
        )

    def _test_init(self) -> None:
        metric = self._test_case_spec.metric
        self.assertEqual(
            set(metric._state_name_to_default.keys()), self._test_case_spec.state_names
        )
        self._test_metric_pickable_hashable(metric)
        self._test_state_dict_load_state_dict(metric)

    def _test_update_and_compute(self) -> None:
        result = None
        test_metric = deepcopy(self._test_case_spec.metric)
        for i in range(self._test_case_spec.num_total_updates):
            # test chainable call
            current_batch_update_kwargs = {
                k: v[i] for k, v in self._test_case_spec.update_kwargs.items()
            }
            result = test_metric.update(**current_batch_update_kwargs).compute()

        final_computation_result = test_metric.compute()
        # compute result from single process should be same as one merged from multiple processes
        assert_result_close(
            final_computation_result,
            self._test_case_spec.compute_result,
            atol=self._test_case_spec.atol,
            rtol=self._test_case_spec.rtol,
        )
        # compute should be idempotent
        assert_result_close(final_computation_result, result)
        self._test_metric_pickable_hashable(test_metric)
        self._test_state_dict_load_state_dict(test_metric)

    def _test_merge_state(self) -> None:
        num_processes = self._test_case_spec.num_processes
        num_total_updates = self._test_case_spec.num_total_updates
        state_names = self._test_case_spec.state_names
        test_metrics: List[Metric] = [
            deepcopy(self._test_case_spec.metric) for i in range(num_processes)
        ]

        # no errors when merge before update, compute result should be the same
        # compared to merge_state is not called
        test_metric_0_copy = deepcopy(test_metrics[0])
        result_before_merge = test_metric_0_copy.update(
            **{k: v[0] for k, v in self._test_case_spec.update_kwargs.items()}
        ).compute()
        test_metrics_copy = deepcopy(test_metrics)
        test_metrics_copy[0].merge_state(test_metrics_copy[1:])
        result_after_merge = (
            test_metrics_copy[0]
            .update(**{k: v[0] for k, v in self._test_case_spec.update_kwargs.items()})
            .compute()
        )
        assert_result_close(result_before_merge, result_after_merge)

        # call merge_state before update
        # update metric 0 and then metric 0 merges metric 1
        test_metric_0_copy = deepcopy(test_metrics[0])
        test_metric_1_copy = deepcopy(test_metrics[1])
        test_metric_0_copy.update(
            **{k: v[0] for k, v in self._test_case_spec.update_kwargs.items()}
        )
        test_metric_0_copy.merge_state([test_metric_1_copy])
        assert_result_close(result_before_merge, test_metric_0_copy.compute())

        # update metric 1 and then metric 0 merges metric 1
        test_metric_0_copy = deepcopy(test_metrics[0])
        test_metric_1_copy = deepcopy(test_metrics[1])
        test_metric_1_copy.update(
            **{k: v[0] for k, v in self._test_case_spec.update_kwargs.items()}
        )
        test_metric_0_copy.merge_state([test_metric_1_copy])
        assert_result_close(result_before_merge, test_metric_0_copy.compute())

        # update, merge, compute
        for i in range(num_processes):
            for j in range(num_total_updates // num_processes):
                metric_i_current_batch_update_kwargs = {
                    k: v[i * num_total_updates // num_processes + j]
                    for k, v in self._test_case_spec.update_kwargs.items()
                }
                test_metrics[i].update(**metric_i_current_batch_update_kwargs).compute()
        test_metrics_unmerged = [deepcopy(metric) for metric in test_metrics]
        final_computation_result = (
            test_metrics[0].merge_state(test_metrics[1:]).compute()
        )
        assert_result_close(
            final_computation_result,
            self._test_case_spec.merge_and_compute_result,
            atol=self._test_case_spec.atol,
            rtol=self._test_case_spec.rtol,
        )

        # input metric states unchanged
        for i in range(1, num_processes):
            self.assert_state_unchanged(
                state_names, test_metrics_unmerged[i], test_metrics[i]
            )

        # compute should be idempotent
        torch.testing.assert_close(
            final_computation_result,
            test_metrics[0].compute(),
            equal_nan=True,
        )
        self._test_metric_pickable_hashable(test_metrics[0])
        self._test_state_dict_load_state_dict(test_metrics[0])

        # metric can still be updated and computed after merged
        test_metrics[0].update(
            **{k: v[0] for k, v in self._test_case_spec.update_kwargs.items()}
        ).compute()

        # merge metrics on different devices
        if torch.cuda.is_available():
            test_metrics_copy = [deepcopy(metric) for metric in test_metrics]
            past_device_type = self._test_case_spec.device
            new_device_type = "cuda" if self._test_case_spec.device == "cpu" else "cuda"
            self.assertEqual(test_metrics_copy[0]._device.type, past_device_type)
            test_metrics_copy[0].to(new_device_type).merge_state(test_metrics_copy[1:])
            for i in range(1, num_processes):
                self.assert_state_unchanged(
                    state_names, test_metrics_copy[i], test_metrics[i]
                )
                self.assertEqual(test_metrics_copy[i]._device.type, past_device_type)
            self.assertEqual(test_metrics_copy[0]._device.type, new_device_type)

    def _test_sync_and_compute(self) -> None:
        lc = pet.LaunchConfig(
            min_nodes=1,
            max_nodes=1,
            nproc_per_node=self._test_case_spec.num_processes,
            run_id=str(uuid.uuid4()),
            rdzv_backend="c10d",
            rdzv_endpoint="localhost:0",
            max_restarts=0,
            monitor_interval=1,
        )
        pet.elastic_launch(lc, entrypoint=self._test_per_process_sync_and_compute)(
            self._test_case_spec,
        )

    @staticmethod
    def _test_per_process_sync_and_compute(
        test_spec: _MetricClassTestCaseSpecs,
    ) -> None:
        init_from_env(device_type="cpu")
        rank = int(os.environ["RANK"])

        metric = clone_metric(test_spec.metric)
        num_total_updates = test_spec.num_total_updates
        num_processes = test_spec.num_processes

        for i in range(num_total_updates // num_processes):
            metric_current_batch_update_kwargs = {
                k: v[rank * num_total_updates // num_processes + i]
                for k, v in test_spec.update_kwargs.items()
            }
            metric.update(**metric_current_batch_update_kwargs).compute()
        final_computation_result = sync_and_compute(metric)
        if rank == 0:
            assert_result_close(
                final_computation_result,
                test_spec.merge_and_compute_result,
                atol=test_spec.atol,
                rtol=test_spec.rtol,
            )

    def assert_state_unchanged(
        self, state_names: Set[str], metric1: Metric, metric2: Metric
    ) -> None:
        for state in state_names:
            assert_result_close(
                getattr(metric1, state),
                getattr(metric2, state),
            )


def assert_result_close(
    # pyre-ignore[2]: There's no restrictions on return types of a specific metric computation
    result: Any,
    # pyre-ignore[2]: There's no restrictions on return types of a specific metric computation
    expected_result: Any,
    atol: float = 1e-8,
    rtol: float = 1e-5,
) -> None:
    tc = unittest.TestCase()
    tc.assertEqual(type(result), type(expected_result))
    if isinstance(result, torch.Tensor):
        tc.assertTrue(isinstance(expected_result, torch.Tensor))
        torch.testing.assert_close(
            result, expected_result, atol=atol, rtol=rtol, equal_nan=True
        )
    elif isinstance(result, Sequence):
        tc.assertTrue(isinstance(expected_result, Sequence))
        tc.assertEqual(len(result), len(expected_result))
        for element, expected_element in zip(result, expected_result):
            assert_result_close(element, expected_element, atol, rtol)
    else:
        # add more supported type to result comparision if needed
        raise ValueError("Compute result comparision is not supported.")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import torch


@torch.inference_mode()
def hit_rate(
    input: torch.Tensor,
    target: torch.Tensor,
    k: Optional[int] = None,
) -> torch.Tensor:
    """
    Compute the hit rate of the correct class among the top predicted classes.
    Its class version is ``torcheval.metrics.HitRate``.

    Args:
        input (Tensor): Predicted unnormalized scores (often referred to as logits) or
            class probabilities of shape (num_samples, num_classes).
        target (Tensor): Ground truth class indices of shape (num_samples,).
        k (int, optional): Number of top predicted classes to be considered.
            If k is None, all classes are considered and a hit rate of 1.0 is returned.

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import hit_rate
        >>> input = torch.tensor([[0.3, 0.1, 0.6], [0.5, 0.2, 0.3], [0.2, 0.1, 0.7], [0.3, 0.3, 0.4]])
        >>> target = torch.tensor([2, 1, 1, 0])
        >>> hit_rate(input, target, k=2)
        tensor([1.0000, 0.0000, 0.0000, 1.0000])
    """
    _hit_rate_input_check(input, target, k)
    if k is None or k >= input.size(dim=-1):
        return input.new_ones(target.size())

    y_score = torch.gather(input, dim=-1, index=target.unsqueeze(dim=-1))
    rank = torch.gt(input, y_score).sum(dim=-1)
    return (rank < k).float()


def _hit_rate_input_check(
    input: torch.Tensor, target: torch.Tensor, k: Optional[int] = None
) -> None:
    if target.ndim != 1:
        raise ValueError(
            f"target should be a one-dimensional tensor, got shape {target.shape}."
        )
    if input.ndim != 2:
        raise ValueError(
            f"input should be a two-dimensional tensor, got shape {input.shape}."
        )
    if input.shape[0] != target.shape[0]:
        raise ValueError(
            "`input` and `target` should have the same minibatch dimension, ",
            f"got shapes {input.shape} and {target.shape}, respectively.",
        )
    if k is not None and k <= 0:
        raise ValueError(f"k should be None or positive, got {k}.")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from collections import defaultdict
from typing import Iterable, TypeVar

import torch
from torcheval.metrics import Metric

TDummySumMetric = TypeVar("TDummySumMetric")


class DummySumMetric(Metric[torch.Tensor]):
    def __init__(self: TDummySumMetric) -> None:
        super().__init__()
        self._add_state("sum", torch.tensor(0.0))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(self: TDummySumMetric, x: torch.Tensor) -> TDummySumMetric:
        self.sum += x
        return self

    @torch.inference_mode()
    def compute(self: TDummySumMetric) -> torch.Tensor:
        return self.sum

    @torch.inference_mode()
    def merge_state(
        self: TDummySumMetric, metrics: Iterable[TDummySumMetric]
    ) -> TDummySumMetric:
        for metric in metrics:
            self.sum += metric.sum.to(self.device)
        return self


TDummySumListStateMetric = TypeVar("TDummySumListStateMetric")


class DummySumListStateMetric(Metric[torch.Tensor]):
    def __init__(self: TDummySumListStateMetric) -> None:
        super().__init__()
        self._add_state("x", [])

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TDummySumListStateMetric, x: torch.Tensor
    ) -> TDummySumListStateMetric:
        self.x.append(x)
        return self

    @torch.inference_mode()
    def compute(self: TDummySumListStateMetric) -> torch.Tensor:
        return sum(tensor.sum() for tensor in self.x)

    @torch.inference_mode()
    def merge_state(
        self: TDummySumListStateMetric, metrics: Iterable[TDummySumListStateMetric]
    ) -> TDummySumListStateMetric:
        for metric in metrics:
            self.x.extend(element.to(self.device) for element in metric.x)
        return self


TDummySumDictStateMetric = TypeVar("TDummySumDictStateMetric")


class DummySumDictStateMetric(Metric[torch.Tensor]):
    def __init__(self: TDummySumDictStateMetric) -> None:
        super().__init__()
        self._add_state("x", defaultdict(lambda: torch.tensor(0.0, device=self.device)))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TDummySumDictStateMetric,
        k: str,
        v: torch.Tensor,
    ) -> TDummySumDictStateMetric:
        self.x[k] += v
        return self

    @torch.inference_mode()
    def compute(self: TDummySumDictStateMetric) -> torch.Tensor:
        return self.x

    @torch.inference_mode()
    def merge_state(
        self: TDummySumDictStateMetric, metrics: Iterable[TDummySumDictStateMetric]
    ) -> TDummySumDictStateMetric:
        for metric in metrics:
            for k in metric.keys():
                self.x[k] += metric.x[k].to(self.device)

        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import logging
import operator
from collections import defaultdict
from functools import reduce
from numbers import Number
from typing import Any, Callable, DefaultDict, Dict, List, Optional, Tuple

import torch
from torch.utils._pytree import PyTree, tree_map

aten = torch.ops.aten  # pyre-fixme need to use the correct type (Module) for this one.


# pyre-ignore [2] we don't care the type in outputs
def _matmul_flop_jit(inputs: Tuple[torch.Tensor], outputs: Tuple[Any]) -> Number:
    """
    Count flops for matmul.
    """
    # Inputs should be a list of length 2.
    # Inputs contains the shapes of two matrices.
    input_shapes = [v.shape for v in inputs]
    assert len(input_shapes) == 2, input_shapes
    assert input_shapes[0][-1] == input_shapes[1][-2], input_shapes
    flop = inputs[0].numel() * input_shapes[-1][-1]
    return flop


# pyre-ignore [2] we don't care the type in outputs
def _addmm_flop_jit(inputs: Tuple[torch.Tensor], outputs: Tuple[Any]) -> Number:
    """
    Count flops for fully connected layers.
    """
    # Count flop for nn.Linear
    # inputs is a list of length 3.
    input_shapes = [v.shape for v in inputs[1:3]]
    # input_shapes[0]: [batch size, input feature dimension]
    # input_shapes[1]: [batch size, output feature dimension]
    assert len(input_shapes[0]) == 2, input_shapes[0]
    assert len(input_shapes[1]) == 2, input_shapes[1]
    batch_size, input_dim = input_shapes[0]
    output_dim = input_shapes[1][1]
    flops = batch_size * input_dim * output_dim
    return flops


# pyre-ignore [2] we don't care the type in outputs
def _bmm_flop_jit(inputs: Tuple[torch.Tensor], outputs: Tuple[Any]) -> Number:
    """
    Count flops for the bmm operation.
    """
    # Inputs should be a list of length 2.
    # Inputs contains the shapes of two tensor.
    assert len(inputs) == 2, len(inputs)
    input_shapes = [v.shape for v in inputs]
    n, c, t = input_shapes[0]
    d = input_shapes[-1][-1]
    flop = n * c * t * d
    return flop


def _conv_flop_count(
    x_shape: List[int],
    w_shape: List[int],
    out_shape: List[int],
    transposed: bool = False,
) -> Number:
    """
    Count flops for convolution. Note only multiplication is
    counted. Computation for addition and bias is ignored.
    Flops for a transposed convolution are calculated as
    flops = (x_shape[2:] * prod(w_shape) * batch_size).
    Args:
        x_shape (list(int)): The input shape before convolution.
        w_shape (list(int)): The filter shape.
        out_shape (list(int)): The output shape after convolution.
        transposed (bool): is the convolution transposed
    Returns:
        int: the number of flops
    """
    batch_size = x_shape[0]
    conv_shape = (x_shape if transposed else out_shape)[2:]
    flop = (
        batch_size
        * reduce(operator.mul, w_shape, 1)
        * reduce(operator.mul, conv_shape, 1)
    )
    return flop


def _conv_flop_jit(
    inputs: Tuple[Any],  # pyre-ignore [2] the inputs can be union of Tensor/bool/Tuple
    outputs: Tuple[torch.Tensor],
) -> Number:
    """
    Count flops for convolution.
    """
    x: torch.Tensor = inputs[0]
    w: torch.Tensor = inputs[1]
    x_shape, w_shape, out_shape = (x.shape, w.shape, outputs[0].shape)
    transposed: bool = inputs[6]

    return _conv_flop_count(
        list(x_shape), list(w_shape), list(out_shape), transposed=transposed
    )


def _transpose_shape(shape: torch.Size) -> List[int]:
    return [shape[1], shape[0]] + list(shape[2:])


# pyre-ignore [2] the inputs can be union of Tensor/bool/Tuple & we don't care about outputs
def _conv_backward_flop_jit(inputs: Tuple[Any], outputs: Tuple[Any]) -> Number:
    grad_out_shape, x_shape, w_shape = [i.shape for i in inputs[:3]]
    output_mask = inputs[-1]
    fwd_transposed = inputs[7]
    flop_count: Number = 0

    if output_mask[0]:
        grad_input_shape = outputs[0].shape
        # pyre-ignore [58] this is actually sum of Number and Number
        flop_count = flop_count + _conv_flop_count(
            grad_out_shape, w_shape, grad_input_shape, not fwd_transposed
        )
    if output_mask[1]:
        grad_weight_shape = outputs[1].shape
        flop_count += _conv_flop_count(
            list(_transpose_shape(x_shape)),
            list(grad_out_shape),
            list(grad_weight_shape),
            fwd_transposed,
        )

    return flop_count


# pyre-ignore [5]
flop_mapping: Dict[Callable[..., Any], Callable[[Tuple[Any], Tuple[Any]], Number]] = {
    aten.mm: _matmul_flop_jit,
    aten.matmul: _matmul_flop_jit,
    aten.addmm: _addmm_flop_jit,
    aten.bmm: _bmm_flop_jit,
    aten.convolution: _conv_flop_jit,
    aten._convolution: _conv_flop_jit,
    aten.convolution_backward: _conv_backward_flop_jit,
    # Add their default to make sure they can be mapped.
    aten.mm.default: _matmul_flop_jit,
    aten.matmul.default: _matmul_flop_jit,
    aten.addmm.default: _addmm_flop_jit,
    aten.bmm.default: _bmm_flop_jit,
    aten.convolution.default: _conv_flop_jit,
    aten._convolution.default: _conv_flop_jit,
    aten.convolution_backward.default: _conv_backward_flop_jit,
}


# pyre-ignore [2, 3] it can be Tuple of anything.
def _normalize_tuple(x: Any) -> Tuple[Any]:
    if not isinstance(x, tuple):
        return (x,)
    return x


# pyre-fixme [13] elem is in __slots__ but obtain an error if initialize it.
class FlopTensor(torch.Tensor):
    elem: torch.Tensor
    flop_counts: DefaultDict[str, DefaultDict[str, int]] = defaultdict(
        lambda: defaultdict(int)
    )
    parents: List[str] = [""]

    __slots__ = ["elem"]

    @staticmethod
    def __new__(cls, elem: torch.Tensor) -> torch.Tensor:
        # The wrapping tensor (FlopTensor) shouldn't hold any
        # memory for the class in question, but it should still
        # advertise the same device as before
        r = torch.Tensor._make_wrapper_subclass(  # type: ignore[attr-defined]
            cls,
            elem.size(),
            strides=elem.stride(),
            storage_offset=elem.storage_offset(),
            # TODO: clone storage aliasing
            dtype=elem.dtype,
            layout=elem.layout,
            device=elem.device,
            requires_grad=elem.requires_grad,
        )
        # ...the real tensor is held as an element on the tensor.
        r.elem = elem
        return r

    def __repr__(self, tensor_contents: Optional[None] = None) -> str:
        if self.grad_fn:
            return f"FlopTensor({self.elem}, grad_fn={self.grad_fn})"
        return f"FlopTensor({self.elem})"

    @classmethod
    def __torch_dispatch__(
        cls,
        func: Callable[..., Any],  # pyre-ignore [2] func can be any func
        types: Tuple[Any],  # pyre-ignore [2]
        args=(),  # pyre-ignore [2]
        kwargs=None,  # pyre-ignore [2]
    ) -> PyTree:
        def unwrap(e: torch.Tensor) -> torch.Tensor:
            return e.elem if isinstance(e, FlopTensor) else e

        # no_dispatch is only needed if you use enable_python_mode.
        # It prevents infinite recursion.
        rs = func(*tree_map(unwrap, args), **tree_map(unwrap, kwargs))
        outs = _normalize_tuple(rs)

        if func in flop_mapping:
            flop_counts = FlopTensor.flop_counts
            flop_count = flop_mapping[func](args, outs)
            for par in FlopTensor.parents:
                # pyre-ignore [58]
                FlopTensor.flop_counts[par][func.__name__] += flop_count
        else:
            logging.debug(f"{func} is not yet supported in FLOPs calculation.")

        def wrap(e: Optional[torch.Tensor]) -> Optional[torch.Tensor]:
            return FlopTensor(e) if isinstance(e, torch.Tensor) else e

        rs = tree_map(wrap, rs)
        return rs


# pyre-ignore [3]
def _create_backwards_push(name: str) -> Callable[..., Any]:
    class PushState(torch.autograd.Function):
        @staticmethod
        def forward(ctx, *args):
            args = tree_map(
                lambda x: x.clone() if isinstance(x, torch.Tensor) else x, args
            )
            if len(args) == 1:
                return args[0]
            return args

        @staticmethod
        def backward(ctx, *grad_outs):
            parents = FlopTensor.parents
            parents.append(name)
            return grad_outs

    # pyre-fixme [16] Not sure why this return error.
    return PushState.apply


# pyre-ignore [3]
def _create_backwards_pop(name: str) -> Callable[..., Any]:
    class PopState(torch.autograd.Function):
        @staticmethod
        def forward(ctx, *args):
            args = tree_map(
                lambda x: x.clone() if isinstance(x, torch.Tensor) else x, args
            )
            if len(args) == 1:
                return args[0]
            return args

        @staticmethod
        def backward(ctx, *grad_outs):
            parents = FlopTensor.parents
            assert parents[-1] == name
            parents.pop()
            return grad_outs

    # pyre-fixme [16] Not sure why this return error.
    return PopState.apply


# pyre-ignore [3] Return a callable function
def _enter_module(name: str) -> Callable[..., Any]:
    # pyre-ignore [2, 3]
    def f(module: torch.nn.Module, inputs: Tuple[Any]):
        parents = FlopTensor.parents
        parents.append(name)
        inputs = _normalize_tuple(inputs)
        out = _create_backwards_pop(name)(*inputs)
        return out

    return f


# pyre-ignore [3] Return a callable function
def _exit_module(name: str) -> Callable[..., Any]:
    # pyre-ignore [2, 3]
    def f(module: torch.nn.Module, inputs: Tuple[Any], outputs: Tuple[Any]):
        parents = FlopTensor.parents
        assert parents[-1] == name
        parents.pop()
        outputs = _normalize_tuple(outputs)
        return _create_backwards_push(name)(*outputs)

    return f


def instrument_module(
    mod: torch.nn.Module,
    all_hooks: List[torch.utils.hooks.RemovableHandle],
    par_name: str,
) -> None:
    for name, module in dict(mod.named_children()).items():
        formatted_name = name
        if par_name != "":
            formatted_name = f"{par_name}.{name}"
        all_hooks.append(
            module.register_forward_pre_hook(_enter_module(formatted_name))
        )
        all_hooks.append(module.register_forward_hook(_exit_module(formatted_name)))
        instrument_module(module, all_hooks, formatted_name)


def start_counting() -> None:
    FlopTensor.parents = [""]
    FlopTensor.flop_counts.clear()



import argparse
import torch

from .utils import get_benchmark_stats, setup_distributed, generate_random_input, InputType, rank_0_print
from typing import Any, Dict, Callable
from torcheval.metrics import Metric, MulticlassF1Score
from torcheval.metrics.toolkit import sync_and_compute
import torch.distributed.launcher as pet
import uuid


### INPUT GENERATION
def generate_update_args_multiclass(num_batches:int, batch_size:int, device: torch.device, num_classes: int) -> Dict[str, Any]:
    input = generate_random_input(InputType.MULTICLASS_LOGIT, device, num_batches, batch_size, num_classes=num_classes)
    target = generate_random_input(InputType.MULTICLASS_LABEL, device, num_batches, batch_size, num_classes=num_classes)
    return {"input": input, "target": target}

### METRIC COMPUTATIONS
def run_metric_computation(metric: Metric, num_batches:int, compute_interval:int, update_kwargs: Dict[str, Any]):
    for batch_idx in range(num_batches):
        metric.update(update_kwargs["input"][batch_idx], update_kwargs["target"][batch_idx])
        if (batch_idx + 1) % compute_interval == 0:
            sync_and_compute(metric)

def main(num_batches, batch_size, compute_interval, num_classes):
    device = setup_distributed()

    rank_0_print(f"Benchmark MulticlassF1Score...")
    metric = MulticlassF1Score().to(device)
    update_kwargs = generate_update_args_multiclass(num_batches, batch_size, device, num_classes)
    get_benchmark_stats(metric, num_batches, compute_interval, update_kwargs)

if __name__ == "__main__":
    print("Benchmark F1Score metrics...")
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-batches", type=int, default=int(10_000))
    parser.add_argument("--batch-size", type=int, default=256)
    parser.add_argument("--compute_interval", type=int, default=100)
    parser.add_argument("--num-processes", type=int, default=4)
    parser.add_argument("--num-classes", type=int, default=10)
    parser.add_argument("--use-pet", action="store_true", default=False)

    args = parser.parse_args()

    if args.use_pet:
        lc = pet.LaunchConfig(
                min_nodes=1,
                max_nodes=1,
                nproc_per_node=args.num_processes,
                run_id=str(uuid.uuid4()),
                rdzv_backend="c10d",
                rdzv_endpoint="localhost:0",
                max_restarts=0,
                monitor_interval=1,
        )
        pet.elastic_launch(lc, entrypoint=main)(
         args.num_batches, args.batch_size, args.compute_interval, args.num_classes
        )
    else:
        main(args.num_batches, args.batch_size, args.compute_interval, args.num_classes)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import typing as T
from datetime import timedelta

import torch

from torch.distributed.constants import default_pg_timeout
from torcheval.utils.device import get_device_from_env
from torcheval.utils.distributed import get_process_group_backend_from_device

_log: logging.Logger = logging.getLogger(__name__)


def _check_dist_env() -> bool:
    """
    Check if all environment variables required to initialize torch.distributed are set
    https://pytorch.org/docs/stable/distributed.html#environment-variable-initialization
    """
    env_required = (
        os.environ.get("MASTER_PORT"),
        os.environ.get("MASTER_ADDR"),
        os.environ.get("WORLD_SIZE"),
        os.environ.get("RANK"),
    )
    return all(env is not None for env in env_required)


def init_from_env(
    device_type: T.Optional[str] = None,
    pg_backend: T.Optional[str] = None,
    pg_timeout: timedelta = default_pg_timeout,
) -> torch.device:
    """
    Utility function that initializes the device and process group, if applicable.
    The global process group is initialized only if:
        - torch distributed is not already initialized
        - the program has been launched on multiple processes
    This is intended as a convenience to include at the beginning of scripts that follow
    a SPMD-style execution model.
    Args:
        device_type (str, optional): Device type to initialize. If None, device will be initialized
                                  based on environment
        pg_backend (str, optional): The process group backend to use. If None, it will use the
                                    defaut process group backend from the device
        pg_timeout (timedelta, optional): Timeout for operations executed against the process
                                          group. Default value equals 30 minutes
    """
    device = torch.device("cpu") if device_type == "cpu" else get_device_from_env()

    if device_type is not None and device.type != device_type:
        raise RuntimeError(
            f"Device type is specified to {device_type} but got {device.type} from env"
        )

    if _check_dist_env():
        if torch.distributed.is_initialized():
            _log.warning(
                "torch.distributed is already initialized. Skipping initializing the process group."
            )
            return device
        pg_backend = (
            pg_backend
            if pg_backend is not None
            else get_process_group_backend_from_device(device)
        )
        torch.distributed.init_process_group(backend=pg_backend, timeout=pg_timeout)
    return device



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Tuple, Union

import torch


@torch.inference_mode()
def binary_binned_precision_recall_curve(
    input: torch.Tensor,
    target: torch.Tensor,
    threshold: Union[int, List[float], torch.Tensor] = 100,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Compute precision recall curve with given thresholds.
    Its class version is ``torcheval.metrics.BinnedPrecisionRecallCurve``.

    Args:
        input: Tensor of label predictions
            It should be probabilities or logits with shape of (n_sample, ).
        target: Tensor of ground truth labels with shape of (n_samples, ).
        threshold:
            a integer representing number of bins, a list of thresholds,
            or a tensor of thresholds.

    Return:
        a tuple of (precision: torch.Tensor, recall: torch.Tensor, thresholds: torch.Tensor)
            precision: Tensor of precision result. Its shape is (n_thresholds + 1, )
            recall: Tensor of recall result. Its shape is (n_thresholds + 1, )
            thresholds: Tensor of threshold. Its shape is (n_thresholds, )

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import binned_precision_recall_curve
        >>> input = torch.tensor([0.2, 0.8, 0.5, 0.9])
        >>> target = torch.tensor([0, 1, 0, 1])
        >>> threshold = 5
        >>> binned_precision_recall_curve(input, target, threshold)
        (tensor([0.5000, 0.6667, 0.6667, 1.0000, 1.0000, 1.0000]),
        tensor([1., 1., 1., 1., 0., 0.]),
        tensor([0.0000, 0.2500, 0.5000, 0.7500, 1.0000]))

        >>> input = torch.tensor([0.2, 0.3, 0.4, 0.5])
        >>> target = torch.tensor([0, 0, 1, 1])
        >>> threshold = torch.tensor([0.0000, 0.2500, 0.7500, 1.0000])
        >>> binned_precision_recall_curve(input, target, threshold)
        (tensor([0.5000, 0.6667, 1.0000, 1.0000, 1.0000]),
        tensor([1., 1., 0., 0., 0.]),
        tensor([0.0000, 0.2500, 0.7500, 1.0000]))
    """
    if isinstance(threshold, int):
        threshold = torch.linspace(0, 1.0, threshold, device=target.device)
    elif isinstance(threshold, list):
        threshold = torch.tensor(threshold, device=target.device)
    num_tp, num_fp, num_fn = _binary_binned_precision_recall_curve_update(
        input, target, threshold
    )
    return _binary_binned_compute_for_each_class(num_tp, num_fp, num_fn, threshold)


def _binary_binned_precision_recall_curve_update(
    input: torch.Tensor,
    target: torch.Tensor,
    threshold: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    _binary_binned_precision_recall_curve_update_input_check(input, target, threshold)
    return _update(input, target, threshold)


@torch.jit.script
def _update(
    input: torch.Tensor,
    target: torch.Tensor,
    threshold: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    pred_label = input >= threshold[:, None]
    num_tp = (pred_label & target).sum(dim=1)
    num_fp = pred_label.sum(dim=1) - num_tp
    num_fn = target.sum() - num_tp
    return num_tp, num_fp, num_fn


def _binary_binned_precision_recall_curve_compute(
    num_tp: torch.Tensor,
    num_fp: torch.Tensor,
    num_fn: torch.Tensor,
    threshold: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    return _binary_binned_compute_for_each_class(num_tp, num_fp, num_fn, threshold)


@torch.jit.script
def _binary_binned_compute_for_each_class(
    num_tp: torch.Tensor,
    num_fp: torch.Tensor,
    num_fn: torch.Tensor,
    threshold: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    # Set precision to 1.0 if all predictions are zeros.
    precision = torch.nan_to_num(num_tp / (num_tp + num_fp), 1.0)
    recall = num_tp / (num_tp + num_fn)

    # The last precision and recall values are 1.0 and 0.0 without a corresponding threshold.
    # This ensures that the graph starts on the y-axis.
    precision = torch.cat([precision, precision.new_ones(1)], dim=0)
    recall = torch.cat([recall, recall.new_zeros(1)], dim=0)

    return precision, recall, threshold


def _binary_binned_precision_recall_curve_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
    threshold: torch.Tensor,
) -> None:
    if input.ndim != 1:
        raise ValueError(
            "input should be a one-dimensional tensor, " f"got shape {input.shape}."
        )

    if target.ndim != 1:
        raise ValueError(
            "target should be a one-dimensional tensor, " f"got shape {target.shape}."
        )

    if input.shape != target.shape:
        raise ValueError(
            "The `input` and `target` should have the same shape, "
            f"got shapes {input.shape} and {target.shape}."
        )

    if (torch.diff(threshold) < 0.0).any():
        raise ValueError("The `threshold` should be a sorted array.")

    if (threshold < 0.0).any() or (threshold > 1.0).any():
        raise ValueError("The values in `threshold` should be in the range of [0, 1].")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import Union

import torch


@torch.inference_mode()
def sum(
    input: torch.Tensor,
    weight: Union[float, torch.Tensor] = 1.0,
) -> torch.Tensor:
    """
    Compute weighted sum. When weight is not provided, it calculates the unweighted sum.
    Its class version is ``torcheval.metrics.Sum``.

    Args:
        input: Tensor of input values.
        weight(optional): Float or Int or Tensor of input weights. It is default to 1.0. If weight is a Tensor, its size should match the input tensor size.
    Raises:
        ValueError: If value of weight is neither a ``float`` nor an ``int`` nor a ``torch.Tensor`` that matches the input tensor size.

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import sum
        >>> sum(torch.tensor([2, 3]))
        tensor(5.)
        >>> sum(torch.tensor([2, 3]), torch.tensor([0.1, 0.6]))
        tensor(2.)
        >>> sum(torch.tensor([2, 3]), 0.5)
        tensor(2.5)
        >>> sum(torch.tensor([2, 3]), 2)
        tensor(10.)
    """
    return _sum_update(input, weight)


def _sum_update(
    input: torch.Tensor, weight: Union[float, int, torch.Tensor]
) -> torch.Tensor:
    if (
        isinstance(weight, float)
        or isinstance(weight, int)
        or (isinstance(weight, torch.Tensor) and input.size() == weight.size())
    ):
        return (input * weight).sum()
    else:
        raise ValueError(
            "Weight must be either a float value or an int value or a tensor that matches the input tensor size. "
            f"Got {weight} instead."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import logging
from copy import deepcopy
from typing import Any, Dict, Iterable, List, Optional, TypeVar, Union

import torch
import torch.distributed as dist
from torcheval.metrics import Metric
from torcheval.utils.distributed import PGWrapper
from typing_extensions import Literal

log: logging.Logger = logging.getLogger(__name__)

_TMetrics = TypeVar("_TMetrics", bound=Iterable[Metric])

# pyre-ignore[3]: compute result can be type Any.
def sync_and_compute(
    metric: Metric,
    process_group: Optional[dist.ProcessGroup] = None,
    recipient_rank: Union[int, Literal["all"]] = 0,
) -> Optional[Any]:
    """
    Sync metric states and returns the ``metric.compute()`` result of
    synced metric on recipient rank. Return ``None`` on other ranks.

    Args:
        metric: The metric object to be synced and computed.
        process_group: The process group on which the metric states are
            gathered. default: ``None`` (the entire world)
        recipient_rank: The destination rank. If string "all" is passed in,
            then all ranks are the destination ranks.

    Example:
        >>> # Assumes world_size of 3.
        >>> # Process group initialization omitted on each rank.
        >>> import torch
        >>> import torch.distributed as dist
        >>> from torcheval.metrics import Max
        >>> max = Max()
        >>> max.update(torch.tensor(dist.get_rank())).compute()
        tensor(0.) # Rank 0
        tensor(1.) # Rank 1
        tensor(2.) # Rank 2
        >>> sync_and_compute(max)
        tensor(2.) # Rank 0
        None # Rank 1
        None # Rank 2
        >>> sync_and_compute(max, recipient_rank="all")
        tensor(2.) # Rank 0
        tensor(2.) # Rank 1
        tensor(2.) # Rank 2
    """
    # Sync metric states to rank 0, compute and broadcast the results
    # when recipient_rank is "all".
    # It uses less memory than syncing metric states to all ranks,
    # since results are usually smaller in size than metric states.
    dst_rank = 0 if recipient_rank == "all" else recipient_rank
    synced_metric = get_synced_metric(metric, process_group, dst_rank)
    compute_result = synced_metric.compute() if synced_metric else None

    if recipient_rank == "all":
        obj_list = [compute_result]
        pg = PGWrapper(process_group)
        if pg.get_rank() == dst_rank:
            pg.broadcast_object_list(obj_list, src=int(dst_rank))
        else:
            pg.broadcast_object_list(obj_list, src=int(dst_rank))
            compute_result = obj_list[0]

    return compute_result


def get_synced_state_dict(
    metric: Metric,
    process_group: Optional[dist.ProcessGroup] = None,
    recipient_rank: Union[int, Literal["all"]] = 0,
) -> Dict[str, Any]:
    """
    Return the state dict of a metric after syncing on recipient_rank.
    Return an empty dict on other ranks.

    Args:
        metric: The metric object to sync and get ``state_dict()``
        process_group: The process group on which the metric states are
            gathered. default: ``None`` (the entire world)
        recipient_rank: The destination rank. If string "all" is passed in,
            then all ranks are the destination ranks.
    Returns:
        state dict of synced metric

    Example:
        >>> # Assumes world_size of 3.
        >>> # Process group initialization omitted on each rank.
        >>> import torch
        >>> import torch.distributed as dist
        >>> from torcheval import Max
        >>> max = Max()
        >>> max.update(torch.tensor(dist.get_rank()))
        >>> get_synced_state_dict(max)
        {"max", tensor(2.)} # Rank 0
        {} # Rank 1
        {} # Rank 2
        >>> get_synced_state_dict(max, recipient_rank="all")
        {"max", tensor(2.)} # Rank 0
        {"max", tensor(2.)} # Rank 1
        {"max", tensor(2.)} # Rank 2
    """
    synced_metric = get_synced_metric(metric, process_group, recipient_rank)
    return synced_metric.state_dict() if synced_metric else {}


def clone_metric(metric: Metric) -> Metric:
    """
    Return a new metric instance which is cloned from the input metric.

    Args:
        metric: The metric object to clone
    Returns:
        A new metric instance from cloning
    """
    return deepcopy(metric)


def clone_metrics(metrics: _TMetrics) -> List[Metric]:
    """
    Return a list of new metric instances which are cloned from the input metrics.

    Args:
        metrics: The metric objects to clone
    Returns:
        A list of metric instances from cloning
    """
    return [clone_metric(metric) for metric in metrics]


def get_synced_metric(
    metric: Metric,
    process_group: Optional[dist.ProcessGroup] = None,
    recipient_rank: Union[int, Literal["all"]] = 0,
) -> Optional[Metric]:
    """
    Returns a metric object on recipient_rank whose internal state
    variables are synced across processes in the process_group.
    Returns ``None`` on non-recipient rank.

    If ``all`` is passed as recipient_rank, all ranks in the
    ``process_group`` are considered as recipient ranks.

    Args:
        metric: The metric object to sync.
        process_group: The process group on which the metric states are
            gathered. default: ``None`` (the entire world)
        recipient_rank: The destination rank. If string "all" is passed in,
            then all ranks are the destination ranks.
    Raises:
        ValueError: when ``recipient_rank`` is not an integer or string
            "all".

    Example:
        >>> # Assumes world_size of 3.
        >>> # Process group initialization omitted on each rank.
        >>> import torch
        >>> import torch.distributed as dist
        >>> from torcheval import Max
        >>> max = Max()
        >>> max.update(torch.tensor(dist.get_rank())).compute()
        tensor(0.) # Rank 0
        tensor(1.) # Rank 1
        tensor(2.) # Rank 2
        >>>  get_synced_metric(max).compute()
        tensor(2.) # Rank 0
        None # Rank 1
        None # Rank 2
         >>>  get_synced_metric(max, recipient_rank=1).compute()
        None # Rank 0
        tensor(2.) # Rank 1
        None # Rank 2
        >>>  get_synced_metric(max, recipient_rank="all").compute()
        tensor(2.) # Rank 0
        tensor(2.) # Rank 1
        tensor(2.) # Rank 2
    """
    if not (isinstance(recipient_rank, int) or recipient_rank == "all"):
        raise ValueError(
            "``recipient_rank`` should be an integer or 'all', "
            f"got {recipient_rank} instead."
        )

    world_size = PGWrapper(process_group).get_world_size()
    if world_size == 1:
        log.warning(
            "World size is 1, and metric is not synced. "
            "``get_synced_metric()`` returns the input metric."
        )
        return metric
    elif world_size == -1:
        log.warning(
            "World size is -1, and current process might not be "
            "in the process group. ``get_synced_metric()`` returns ``None``."
        )
        return None
    if world_size <= 1:
        raise RuntimeError(
            f"Unexpected world_size {world_size} is seen when syncing metrics!"
        )

    gathered_metric_list = _sync_metric_object(
        metric,
        # pyre-ignore[6]: expect `ProcessGroup` but got `Optional[ProcessGroup]`.
        process_group if process_group else dist.group.WORLD,
        # pyre-ignore[6]: expect `Union[int, Literal["all"]`, got `Union[int, str]`.
        recipient_rank,
        world_size,
    )

    if gathered_metric_list is None:
        return None
    return (
        clone_metric(gathered_metric_list[0])
        .to(metric.device)
        .merge_state(gathered_metric_list[1:])
    )


def _sync_metric_object(
    metric: Metric,
    process_group: dist.ProcessGroup,
    recipient_rank: Union[int, Literal["all"]],
    world_size: int,
) -> Optional[List[Metric]]:
    metric._prepare_for_merge_state()
    gathered_metric_list = (
        [None] * world_size
        if recipient_rank == "all" or dist.get_rank() == recipient_rank
        else None
    )
    if isinstance(recipient_rank, int):
        dist.gather_object(
            metric,
            gathered_metric_list,
            dst=recipient_rank,
            group=process_group,
        )
    elif recipient_rank == "all":
        dist.all_gather_object(gathered_metric_list, metric, group=process_group)
    # pyre-ignore[7]: Expected `Optional[List[Metric[typing.Any]]]` but got `Optional[List[None]]`
    return gathered_metric_list


def reset_metrics(metrics: _TMetrics) -> _TMetrics:
    """
    Reset input metrics and returns the reset collection back to users.

    Args:
        metrics: The metrics to be reset

    Example:
        >>> from torcheval.metrics import Max, Min
        >>> max = Max()
        >>> min = Min()
        >>> max.update(torch.tensor(1)).compute()
        >>> min.update(torch.tensor(2)).compute()
        >>> max, min = reset_metrics((max, min))
        >>> max.compute()
        tensor(0.)
        >>> min.compute()
        tensor(0.)
    """

    for metric in metrics:
        metric.reset()
    return metrics


def to_device(
    metrics: _TMetrics, device: torch.device, *args: Any, **kwargs: Any
) -> _TMetrics:
    """
    Moves input metrics to the target device and returns the moved metrics back to users.

    Args:
        metrics: The metrics to be moved to the device
        device: the device to move te metrics to
        *args: Variadic arguments forwarded to ``Metric.to``
        **kwargs: Named arguments forwarded to ``Metric.to``

    Example:
        >>> from torcheval.metrics import Max, Min
        >>> max = Max()
        >>> min = Min()
        >>> max, min = to_device((max, min), torch.device("cuda"))
        >>> max.device
        torch.device("cuda")
        >>> min.device
        torch.device("cuda")
    """
    for metric in metrics:
        metric.to(device, *args, **kwargs)
    return metrics



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import logging

from typing import List, Optional, Tuple

import torch
import torch.nn.functional as F

"""
This file contains binary_precision_recall_curve and multiclass_precision_recall_curve functions.
"""


@torch.inference_mode()
def binary_precision_recall_curve(
    input: torch.Tensor,
    target: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Compute precision recall curve, which is precision-recall pair with corresponding thresholds,
        for binary classification tasks.
    Its class version is ``torcheval.metrics.BinaryPrecisionRecallCurve``.

    Args:
        input: Tensor of label predictions
            It should be probabilities or logits with shape of (n_sample, ).
        target: Tensor of ground truth labels with shape of (n_samples, ).

    Return:
        a tuple of (precision: torch.Tensor, recall: torch.Tensor, thresholds: torch.Tensor)
            precision: Tensor of precision result. Its shape is (n_thresholds + 1, )
            recall: Tensor of recall result. Its shape is (n_thresholds + 1, )
            thresholds: Tensor of threshold. Its shape is (n_thresholds, )

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import binary_precision_recall_curve
        >>> input = torch.tensor([0.1, 0.5, 0.7, 0.8])
        >>> target = torch.tensor([0, 0, 1, 1])
        >>> binary_precision_recall_curve(input, target)
        (tensor([1., 1., 1.]),
        tensor([1.0000, 0.5000, 0.0000]),
        tensor([0.7000, 0.8000]))
    """
    _binary_precision_recall_curve_update(input, target)
    return _binary_precision_recall_curve_compute(input, target)


def _binary_precision_recall_curve_update(
    input: torch.Tensor,
    target: torch.Tensor,
) -> None:
    _binary_precision_recall_curve_update_input_check(input, target)


def _binary_precision_recall_curve_compute(
    input: torch.Tensor,
    target: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    precision, recall, threshold = _compute_for_each_class(input, target, 1)

    return precision, recall, threshold


def _binary_precision_recall_curve_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
) -> None:
    if input.ndim != 1:
        raise ValueError(
            "input should be a one-dimensional tensor, " f"got shape {input.shape}."
        )

    if target.ndim != 1:
        raise ValueError(
            "target should be a one-dimensional tensor, " f"got shape {target.shape}."
        )

    if input.shape != target.shape:
        raise ValueError(
            "The `input` and `target` should have the same shape, "
            f"got shapes {input.shape} and {target.shape}."
        )


@torch.inference_mode()
def multiclass_precision_recall_curve(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int] = None,
) -> Tuple[List[torch.Tensor], List[torch.Tensor], List[torch.Tensor]]:
    """
    Compute precision recall curve, which is precision-recall pair with corresponding thresholds,
        for multi-class classification tasks.
    Its class version is ``torcheval.metrics.MulticlassPrecisionRecallCurve``.

    Args:
        input: Tensor of label predictions
            It should be probabilities or logits with shape of (n_sample, n_class).
        target: Tensor of ground truth labels with shape of (n_samples, ).
        num_classes (Optional):
            Number of classes. Set to the second dimension of the input if num_classes is None.

    Return:
        a tuple of (precision: List[torch.Tensor], recall: List[torch.Tensor], thresholds: List[torch.Tensor])
            precision: List of precision result. Each index indicates the result of a class.
            recall: List of recall result. Each index indicates the result of a class.
            thresholds: List of threshold. Each index indicates the result of a class.

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import multiclass_precision_recall_curve
        >>> input = torch.tensor([[0.1, 0.1, 0.1, 0.1], [0.5, 0.5, 0.5, 0.5], [0.7, 0.7, 0.7, 0.7], [0.8, 0.8, 0.8, 0.8]])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> precision_recall_curve(input, target, num_classes=4)
        ([tensor([0.2500, 0.0000, 0.0000, 0.0000, 1.0000]),
        tensor([0.3333, 0.0000, 0.0000, 1.0000]),
        tensor([0.5000, 0.0000, 1.0000]),
        tensor([1., 1.])],
        [tensor([1., 0., 0., 0., 0.]),
        tensor([1., 0., 0., 0.]),
        tensor([1., 0., 0.]),
        tensor([1., 0.])],
        [tensor([0.1000, 0.5000, 0.7000, 0.8000]),
        tensor([0.5000, 0.7000, 0.8000]),
        tensor([0.7000, 0.8000]),
        tensor([0.8000])])
    """
    if num_classes is None and input.ndim == 2:
        num_classes = input.shape[1]
    _multiclass_precision_recall_curve_update(input, target, num_classes)
    return _multiclass_precision_recall_curve_compute(input, target, num_classes)


def _multiclass_precision_recall_curve_update(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
) -> None:
    _multiclass_precision_recall_curve_update_input_check(input, target, num_classes)


def _multiclass_precision_recall_curve_compute(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
) -> Tuple[List[torch.Tensor], List[torch.Tensor], List[torch.Tensor]]:
    if num_classes is None and input.ndim == 2:
        num_classes = input.shape[1]
    num_unique_classes = torch.unique(target)
    if len(num_unique_classes) != num_classes:
        logging.warning(
            "Warning: Some classes do not exist in the target. Precision Recall Curve for these classes will be straight lines from (1, 0) to (0, 1)."
        )

    precisions, recalls, thresholds = [], [], []
    assert isinstance(num_classes, int)
    for class_idx in range(num_classes):
        precision, recall, threshold = _compute_for_each_class(
            input[:, class_idx], target, class_idx
        )
        precisions.append(precision)
        recalls.append(recall)
        thresholds.append(threshold)
    return (
        precisions,
        recalls,
        thresholds,
    )


def _multiclass_precision_recall_curve_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
) -> None:
    if input.size(0) != target.size(0):
        raise ValueError(
            "The `input` and `target` should have the same first dimension, "
            f"got shapes {input.shape} and {target.shape}."
        )

    if target.ndim != 1:
        raise ValueError(
            "target should be a one-dimensional tensor, " f"got shape {target.shape}."
        )

    if not (input.ndim == 2 and (num_classes is None or input.shape[1] == num_classes)):
        raise ValueError(
            f"input should have shape of (num_sample, num_classes), "
            f"got {input.shape} and num_classes={num_classes}."
        )


@torch.jit.script
def _compute_for_each_class(
    input: torch.Tensor,
    target: torch.Tensor,
    pos_label: int,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    threshold, indices = input.sort(descending=True)
    mask = F.pad(threshold.diff(dim=0) != 0, [0, 1], value=1.0)
    num_tp = (target[indices] == pos_label).cumsum(0)[mask]
    num_fp = (1 - (target[indices] == pos_label).long()).cumsum(0)[mask]
    precision = num_tp / (num_tp + num_fp)
    recall = num_tp / num_tp[-1]

    # Remove redundant thresholds that result in a recall of 1.0.
    last_ind = torch.searchsorted(num_tp, num_tp[-1])
    sl = slice(last_ind + 1)
    precision = precision[sl].flip(0)
    recall = recall[sl].flip(0)
    threshold = threshold[mask][sl].flip(0)

    # The last precision and recall values are 1.0 and 0.0 without a corresponding threshold.
    # This ensures that the graph starts on the y-axis.
    precision = torch.cat([precision, precision.new_ones(1)])
    recall = torch.cat([recall, recall.new_zeros(1)])

    # If all recalls are NaN, the curve will be a straight line from (1, 0) to (0, 1).
    if torch.isnan(recall[0]):
        precision = torch.tensor([0.0, 1.0], device=precision.device)
        recall = torch.tensor([1.0, 0.0], device=recall.device)

    return precision, recall, threshold



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from torcheval.metrics.functional.regression.mean_squared_error import (
    mean_squared_error,
)
from torcheval.metrics.functional.regression.r2_score import r2_score

__all__ = ["mean_squared_error", "r2_score"]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from torcheval.metrics.functional.aggregation import sum
from torcheval.metrics.functional.classification import (
    binary_accuracy,
    binary_auroc,
    binary_binned_precision_recall_curve,
    binary_precision_recall_curve,
    multiclass_accuracy,
    multiclass_f1_score,
    multiclass_precision,
    multiclass_precision_recall_curve,
    multiclass_recall,
    multilabel_accuracy,
)
from torcheval.metrics.functional.ranking import hit_rate, reciprocal_rank
from torcheval.metrics.functional.regression import mean_squared_error, r2_score

__all__ = [
    "binary_auroc",
    "binary_accuracy",
    "binary_precision_recall_curve",
    "binary_binned_precision_recall_curve",
    "hit_rate",
    "mean",
    "mean_squared_error",
    "multiclass_accuracy",
    "multiclass_f1_score",
    "multiclass_precision",
    "multiclass_precision_recall_curve",
    "multiclass_recall",
    "multilabel_accuracy",
    "r2_score",
    "reciprocal_rank",
    "sum",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Optional, Tuple

import torch


@torch.inference_mode()
def multiclass_recall(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int] = None,
    average: Optional[str] = "micro",
) -> torch.Tensor:
    """
    Compute recall score, which is calculated as the ratio between the number of
    true positives (TP) and the total number of actual positives (TP + FN).
    Its class version is ``torcheval.metrics.MultiClassRecall``.

    Args:
        input: Tensor of label predictions
            It could be the predicted labels, with shape of (n_sample, ).
            It could also be probabilities or logits with shape of (n_sample, n_class).
            ``torch.argmax`` will be used to convert input into predicted labels.
        target: Tensor of ground truth labels with shape of (n_sample, ).
        num_classes:
            Number of classes.
        average:
            - ``'micro'``[default]:
                Calculate the metrics globally, by using the total true positives and false
                negatives across all classes.
            - ``'macro'``:
                Calculate metrics for each class separately, and return their unweighted
                mean. Classes with 0 true and predicted instances are ignored.
            - ``'weighted'``:
                Calculate metrics for each class separately, and return their average weighted
                by the number of instances for each class in the ``target`` tensor. Classes with
                0 true and predicted instances are ignored.
            - ``None``:
                Calculate the metric for each class separately, and return
                the metric for every class.

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional.classification import multiclass_recall
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> multiclass_recall(input, target)
        tensor(0.5000)
        >>> multiclass_recall(input, target, average=None, num_classes=4)
        tensor([1., 0., 0., 1.])
        >>> multiclass_recall(input, target, average="macro", num_classes=4)
        tensor(0.5000)
        >>> input = torch.tensor([[0.9, 0.1, 0, 0], [0.1, 0.2, 0.4, 0.3], [0, 1.0, 0, 0], [0, 0, 0.2, 0.8]])
        >>> multiclass_recall(input, target)
        tensor(0.5000)
    """
    _recall_param_check(num_classes, average)
    num_tp, num_labels, num_predictions = _recall_update(
        input, target, num_classes, average
    )
    return _recall_compute(num_tp, num_labels, num_predictions, average)


def _recall_update(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
    average: Optional[str],
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    _recall_update_input_check(input, target, num_classes)

    if input.ndim == 2:
        input = torch.argmax(input, dim=1)

    if average == "micro":
        num_tp = (input == target).sum()
        num_labels = target.new_tensor(target.numel())
        num_predictions = num_labels
        return num_tp, num_labels, num_predictions

    assert isinstance(
        num_classes, int
    ), f"`num_classes` must be an integer, but received {num_classes}."
    input = input.to(dtype=torch.float32)
    target = target.to(dtype=torch.float32)
    num_tp = torch.histc(
        input[input == target], bins=num_classes, min=0, max=num_classes - 1
    )
    num_labels = torch.histc(target, bins=num_classes, min=0, max=num_classes - 1)
    num_predictions = torch.histc(input, bins=num_classes, min=0, max=num_classes - 1)
    return num_tp, num_labels, num_predictions


def _recall_compute(
    num_tp: torch.Tensor,
    num_labels: torch.Tensor,
    num_predictions: torch.Tensor,
    average: Optional[str],
) -> torch.Tensor:
    if average in ("macro", "weighted"):
        # Ignore classes which have no samples in `target` and `input`
        mask = (num_labels != 0) | (num_predictions != 0)
        num_tp = num_tp[mask]

    recall = num_tp / num_labels

    isnan_class = torch.isnan(recall)
    if isnan_class.any():
        nan_classes = isnan_class.nonzero(as_tuple=True)[0]
        logging.warning(
            f"One or more NaNs identified, as no ground-truth instances of "
            f"{nan_classes.tolist()} have been seen. These have been converted to zero."
        )
        recall = torch.nan_to_num(recall)

    if average == "micro":
        return recall
    elif average == "macro":
        return recall.mean()
    elif average == "weighted":
        weights = num_labels[mask] / num_labels.sum()
        return (recall * weights).sum()
    else:  # average is None
        return recall


def _recall_param_check(num_classes: Optional[int], average: Optional[str]) -> None:
    average_options = ("micro", "macro", "weighted", None)
    if average not in average_options:
        raise ValueError(
            f"`average` was not in the allowed values of {average_options}, "
            f"got {average}."
        )
    if average != "micro" and (num_classes is None or num_classes <= 0):
        raise ValueError(
            f"`num_classes` should be a positive number when average={average}, "
            f"got num_classes={num_classes}."
        )


def _recall_update_input_check(
    input: torch.Tensor, target: torch.Tensor, num_classes: Optional[int]
) -> None:
    if input.size(0) != target.size(0):
        raise ValueError(
            f"The `input` and `target` should have the same first dimension, "
            f"got shapes {input.shape} and {target.shape}."
        )
    if target.ndim != 1:
        raise ValueError(
            f"`target` should be a one-dimensional tensor, got shape {target.shape}."
        )
    if input.ndim != 1 and not (
        input.ndim == 2 and (num_classes is None or input.shape[1] == num_classes)
    ):
        raise ValueError(
            f"`input` should have shape (num_samples,) or (num_samples, num_classes), "
            f"got {input.shape}."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional, Tuple

import torch


@torch.inference_mode()
def binary_accuracy(
    input: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.5,
) -> torch.Tensor:
    """
    Compute binary accuracy score, which is the frequency of input matching target.
    Its functional version is ``torcheval.metrics.functional.binary_accuracy``.

    Args:
        input: Tensor of label predictions with shape of (n_sample,).
            ``torch.where(input < threshold, 0, 1)`` will be applied to the input.
        target: Tensor of ground truth labels with shape of (n_sample,).
        threshold [default: 0.5]: Threshold for converting input into predicted labels for each sample.
            ``torch.where(input < threshold, 0, 1)`` will be applied to the ``input``.
    Example:
        >>> import torch
        >>> from torcheval.metrics import BinaryAccuracy
        >>> metric = BinaryAccuracy()
        >>> input = torch.tensor([0, 0, 1, 1])
        >>> target = torch.tensor([1, 0, 1, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.75)  # 3 / 4

        >>> metric = BinaryAccuracy(threshold=0.7)
        >>> input = torch.tensor([0, 0.2, 0.6, 0.7])
        >>> target = torch.tensor([1, 0, 1, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5)  # 2 / 4
    """

    num_correct, num_total = _binary_accuracy_update(input, target, threshold)
    return _accuracy_compute(num_correct, num_total, "micro")


@torch.inference_mode()
def multiclass_accuracy(
    input: torch.Tensor,
    target: torch.Tensor,
    average: Optional[str] = "micro",
    num_classes: Optional[int] = None,
) -> torch.Tensor:
    """
    Compute accuracy score, which is the frequency of input matching target.
    Its class version is ``torcheval.metrics.MultiClassAccuracy``.

    Args:
        input: Tensor of label predictions
            It could be the predicted labels, with shape of (n_sample, ).
            It could also be probabilities or logits with shape of (n_sample, n_class).
            ``torch.argmax`` will be used to convert input into predicted labels.
        target: Tensor of ground truth labels with shape of (n_sample, ).
        average:
            - ``'micro'``[default]:
                Calculate the metrics globally.
            - ``'macro'``:
                Calculate metrics for each class separately, and return their unweighted
                mean. Classes with 0 true instances are ignored.
            - ``None``:
                Calculate the metric for each class separately, and return
                the metric for every class.
                NaN is returned if a class has no sample in ``target``.
        num_classes:
            Number of classes. Required for ``'macro'`` and ``None`` average methods.

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import multiclass_accuracy
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> multiclass_accuracy(input, target)
        tensor(0.5)
        >>> multiclass_accuracy(input, target, average=None, num_classes=4)
        tensor([1., 0., 0., 1.])
        >>> multiclass_accuracy(input, target, average="macro", num_classes=4)
        tensor(0.5)
        >>> input = torch.tensor([[0.9, 0.1, 0, 0], [0.1, 0.2, 0.4, 0,3], [0, 1.0, 0, 0], [0, 0, 0.2, 0.8]])
        >>> multiclass_accuracy(input, target)
        tensor(0.5)
    """

    _accuracy_param_check(average, num_classes)
    num_correct, num_total = _multiclass_accuracy_update(
        input, target, average, num_classes
    )
    return _accuracy_compute(num_correct, num_total, average)


@torch.inference_mode()
def multilabel_accuracy(
    input: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.5,
    criteria: str = "exact_match",
) -> torch.Tensor:
    """
    Compute multilabel accuracy score, which is the frequency of input matching target.
    Its class version is ``torcheval.metrics.MultilabelAccuracy``.

    Args:
        input: Tensor of label predictions with shape of (n_sample, n_class).
            `torch.where(input < threshold, 0, 1)`` will be applied to the ``input``.
        target: Tensor of ground truth labels with shape of (n_sample, n_class).
        threshold: Threshold for converting input into predicted labels for each sample.
        criteria:
        - ``'exact_match'``[default]:
            The set of labels predicted for a sample must exactly match the corresponding
            set of labels in target. Also known as subset accuracy.
        - ``'hamming'``:
            Fraction of correct labels over total number of labels.

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import multilabel_accuracy
        >>> input = torch.tensor([[0, 1], [1, 1], [0, 0], [0, 1]])
        >>> target = torch.tensor([[0, 1], [1, 0], [0, 0], [1, 1]])
        >>> multilabel_accuracy(input, target)
        tensor(0.5)  # 2 / 4

        >>> input = torch.tensor([[0, 1], [1, 1], [0, 0], [0, 1]])
        >>> target = torch.tensor([[0, 1], [1, 0], [0, 0], [1, 1]])
        >>> multilabel_accuracy(input, target, criteria="hamming")
        tensor(0.75)  # 6 / 8
    """

    _multilabel_accuracy_param_check(criteria)
    num_correct, num_total = _multilabel_accuracy_update(
        input, target, threshold, criteria
    )
    return _accuracy_compute(num_correct, num_total, "micro")


def _multiclass_accuracy_update(
    input: torch.Tensor,
    target: torch.Tensor,
    average: Optional[str],
    num_classes: Optional[int],
) -> Tuple[torch.Tensor, torch.Tensor]:

    _accuracy_update_input_check(input, target, num_classes)

    if input.ndim == 2:
        input = torch.argmax(input, dim=1)

    if average == "micro":
        num_correct = (input == target).sum()
        num_total = torch.tensor(target.shape[0])
        return num_correct, num_total

    # pyre-ignore[6]: expect int got Optional[int] for num_classes
    num_correct = input.new_zeros((num_classes)).scatter_(
        0, target, (target == input).long(), reduce="add"
    )
    # pyre-ignore[6]: expect int got Optional[int] for num_classes
    num_total = target.new_zeros((num_classes)).scatter_(0, target, 1, reduce="add")
    return num_correct, num_total


@torch.jit.script
def _accuracy_compute(
    num_correct: torch.Tensor,
    num_total: torch.Tensor,
    average: Optional[str],
) -> torch.Tensor:
    if isinstance(average, str) and average == "macro":
        mask = num_total != 0
        return (num_correct[mask] / num_total[mask]).mean()
    else:
        return num_correct / num_total


def _accuracy_param_check(
    average: Optional[str],
    num_classes: Optional[int] = None,
) -> None:
    average_options = ("micro", "macro", "none", None)
    if average not in average_options:
        raise ValueError(
            f"`average` was not in the allowed value of {average_options}, got {average}."
        )
    if average != "micro" and (num_classes is None or num_classes <= 0):
        raise ValueError(
            f"num_classes should be a positive number when average={average}."
            f" Got num_classes={num_classes}."
        )


def _accuracy_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
) -> None:
    if input.size(0) != target.size(0):
        raise ValueError(
            "The `input` and `target` should have the same first dimension, "
            f"got shapes {input.shape} and {target.shape}."
        )

    if target.ndim != 1:
        raise ValueError(
            f"target should be a one-dimensional tensor, got shape {target.shape}."
        )

    if not input.ndim == 1 and not (
        input.ndim == 2 and (num_classes is None or input.shape[1] == num_classes)
    ):
        raise ValueError(
            "input should have shape of (num_sample,) or (num_sample, num_classes), "
            f"got {input.shape}."
        )


def _binary_accuracy_update(
    input: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.5,
) -> Tuple[torch.Tensor, torch.Tensor]:

    _binary_accuracy_update_input_check(input, target)

    input = torch.where(input < threshold, 0, 1)

    num_correct = (input == target).sum()
    num_total = target.new_tensor(target.shape[0])
    return num_correct, num_total


def _binary_accuracy_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
) -> None:
    if input.shape != target.shape:
        raise ValueError(
            "The `input` and `target` should have the same dimensions, "
            f"got shapes {input.shape} and {target.shape}."
        )
    if target.ndim != 1:
        raise ValueError(
            f"target should be a one-dimensional tensor, got shape {target.shape}."
        )


def _multilabel_accuracy_update(
    input: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.5,
    criteria: str = "exact_match",
) -> Tuple[torch.Tensor, torch.Tensor]:
    _multilabel_accuracy_update_input_check(input, target)
    return _multilabel_update(input, target, threshold, criteria)


@torch.jit.script
def _multilabel_update(
    input: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.5,
    criteria: str = "exact_match",
) -> Tuple[torch.Tensor, torch.Tensor]:
    input = torch.where(input < threshold, 0, 1)

    if criteria == "exact_match":
        num_correct = torch.all(input == target, dim=1).sum()
        num_total = torch.tensor(target.shape[0])
        return num_correct, num_total

    # Hamming
    num_correct = (input == target).sum()
    num_total = torch.tensor(target.numel(), device=target.device)
    return num_correct, num_total


def _multilabel_accuracy_param_check(
    criteria: str,
) -> None:
    criteria_options = ("exact_match", "hamming")
    if criteria not in criteria_options:
        raise ValueError(
            f"`criteria` was not in the allowed value of {criteria_options}, got {criteria}."
        )


def _multilabel_accuracy_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
) -> None:
    if input.shape != target.shape:
        raise ValueError(
            "The `input` and `target` should have the same dimensions, "
            f"got shapes {input.shape} and {target.shape}."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[6]: expect int got Optional[int] for num_classes

import logging
from typing import Optional, Tuple

import torch


@torch.inference_mode()
def multiclass_f1_score(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int] = None,
    average: Optional[str] = "micro",
) -> torch.Tensor:
    """
    Compute f1 score, which is defined as the harmonic mean of precision and recall.
    We convert NaN to zero when f1 score is NaN. This happens when either precision
    or recall is NaN or when both precision and recall are zero.
    Its class version is ``torcheval.metrics.MultiClassF1Score``.

    Args:
        input: Tensor of label predictions.
            It could be the predicted labels, with shape of (n_sample, ).
            It could also be probabilities or logits with shape of (n_sample, n_class).
            ``torch.argmax`` will be used to convert input into predicted labels.
        target: Tensor of ground truth labels with shape of (n_sample, ).
        num_classes:
            Number of classes.
        average:
            - ``'micro'``[default]:
                Calculate the metrics globally.
            - ``'macro'``:
                Calculate metrics for each class separately, and return their unweighted mean.
                Classes with 0 true and predicted instances are ignored.
            - ``'weighted'``"
                Calculate metrics for each class separately, and return their weighted sum.
                Weights are defined as the proportion of occurrences of each class in "target".
                Classes with 0 true and predicted instances are ignored.
            - ``None``:
                Calculate the metric for each class separately, and return
                the metric for every class.

    Example:
        >>> import torch
        >>> from torcheval.metrics.functional import multiclass_f1_score
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> multiclass_f1_score(input, target, num_classes=4)
        tensor(0.5000)

        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> multiclass_f1_score(input, target, num_classes=4, average=None)
        tensor([1., 0., 0., 1.])

        >>> input = torch.tensor([0, 0, 1, 1, 1])
        >>> target = torch.tensor([0, 0, 0, 0, 1])
        >>> multiclass_f1_score(input, target, num_classes=2, average="macro")
        tensor(0.5833)

        >>> input = torch.tensor([[0.9, 0.1, 0, 0], [0.1, 0.2, 0.4, 0.3], [0, 1.0, 0, 0], [0, 0, 0.2, 0.8]])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> multiclass_f1_score(input, target, num_classes=4)
        tensor(0.5)
    """
    _f1_score_param_check(num_classes, average)
    num_tp, num_label, num_prediction = _f1_score_update(
        input, target, num_classes, average
    )
    return _f1_score_compute(num_tp, num_label, num_prediction, average)


def _f1_score_update(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
    average: Optional[str],
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    _f1_score_update_input_check(input, target, num_classes)
    return _update(input, target, num_classes, average)


@torch.jit.script
def _update(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
    average: Optional[str],
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    if input.ndim == 2:
        input = torch.argmax(input, dim=1)

    if isinstance(average, str) and average == "micro":
        num_tp = (input == target).sum()
        num_label = torch.tensor(target.shape[0], device=target.device)
        num_prediction = num_label
        return num_tp, num_label, num_prediction

    # Add this line to bypass torch jit datatype checking
    assert isinstance(num_classes, int)
    num_label = torch.zeros(num_classes, device=target.device).scatter_(
        0, target, 1, reduce="add"
    )
    num_prediction = torch.zeros(num_classes, device=target.device).scatter_(
        0, input, 1, reduce="add"
    )
    num_tp = torch.zeros(num_classes, device=target.device).scatter_(
        0, target[input == target], 1, reduce="add"
    )
    return num_tp, num_label, num_prediction


def _f1_score_compute(
    num_tp: torch.Tensor,
    num_label: torch.Tensor,
    num_prediction: torch.Tensor,
    average: Optional[str],
) -> torch.Tensor:
    # Check if all classes exist in either ``target``.
    num_label_is_zero = num_label == 0
    if num_label_is_zero.any():
        logging.warning(
            "Warning: Some classes do not exist in the target. F1 scores for these classes will be cast to zeros."
        )

    if average in ("macro", "weighted"):
        # Ignore the class that has no samples in both ``input`` and `target`.
        mask = (~num_label_is_zero) | (num_prediction != 0)
        num_tp, num_label, num_prediction = (
            num_tp[mask],
            num_label[mask],
            num_prediction[mask],
        )

    precision = num_tp / num_prediction
    recall = num_tp / num_label
    f1 = 2 * precision * recall / (precision + recall)

    # Convert NaN to zero when f1 score is NaN. This happens when either precision or recall is NaN or when both precision and recall are zero.
    f1 = torch.nan_to_num(f1)

    if average == "micro":
        return f1
    elif average == "macro":
        return f1.mean()
    elif average == "weighted":
        return (f1 * (num_label[mask] / num_label.sum())).sum()
    else:  # average is None
        return f1


def _f1_score_param_check(
    num_classes: Optional[int],
    average: Optional[str],
) -> None:
    average_options = ("micro", "macro", "weighted", None)
    if average not in average_options:
        raise ValueError(
            f"`average` was not in the allowed value of {average_options}, got {average}."
        )
    if average != "micro" and (num_classes is None or num_classes <= 0):
        raise ValueError(
            f"num_classes should be a positive number when average={average}, "
            f"got num_classes={num_classes}."
        )


def _f1_score_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
    num_classes: Optional[int],
) -> None:
    if input.size(0) != target.size(0):
        raise ValueError(
            "The `input` and `target` should have the same first dimension, "
            f"got shapes {input.shape} and {target.shape}."
        )

    if target.ndim != 1:
        raise ValueError(
            "target should be a one-dimensional tensor, " f"got shape {target.shape}."
        )

    if not input.ndim == 1 and not (
        input.ndim == 2 and (num_classes is None or input.shape[1] == num_classes)
    ):
        raise ValueError(
            f"input should have shape of (num_sample,) or (num_sample, num_classes), "
            f"got {input.shape}."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional, Tuple

import torch


@torch.inference_mode()
def mean_squared_error(
    input: torch.Tensor,
    target: torch.Tensor,
    sample_weight: Optional[torch.Tensor] = None,
    multioutput: str = "uniform_average",
) -> torch.Tensor:
    """
    Compute Mean Squared Error, which is the mean of squared error of `input` and `target`
    Its class version is ``torcheval.metrics.MeanSquaredError``.

    Args:
        input: Tensor of predicted values with shape of (n_sample, n_output).
        target: Tensor of ground truth values with shape of (n_sample, n_output).
        sample_weight (Optional):
            Tensor of sample weights with shape of (n_sample, ). Defaults to None.
        multioutput (Optional):
            - ``'uniform_average'``[default]:
                Return scores of all outputs are averaged with uniform weight.
            - ``'raw_values'``:
                Return a full set of scores.
    Raises:
        ValueError:
            - If value of multioutput does not exist in (``raw_values``, ``uniform_average``).
            - If the dimension of `input` or `target` is not 1D or 2D.
            - If the `input` and `target` do not have the same size.
            - If the first dimension of `input`, `target` and `sample_weight` are not the same.

    Example:
        >>> import torch
        >>> from torcheval.metrics.function import mean_squared_error
        >>> input = torch.tensor([0.9, 0.5, 0.3, 0.5])
        >>> target = torch.tensor([0.5, 0.8, 0.2, 0.8])
        >>> mean_squared_error(input, target)
        tensor(0.0875)

        >>> input = torch.tensor([[0.9, 0.5], [0.3, 0.5]])
        >>> target = torch.tensor([[0.5, 0.8], [0.2, 0.8]])
        >>> mean_squared_error(input, target)
        tensor(0.0875)

        >>> input = torch.tensor([[0.9, 0.5], [0.3, 0.5]])
        >>> target = torch.tensor([[0.5, 0.8], [0.2, 0.8]])
        >>> mean_squared_error(input, target, multioutput="raw_values")
        tensor([0.0850, 0.0900])

        >>> input = torch.tensor([[0.9, 0.5], [0.3, 0.5]])
        >>> target = torch.tensor([[0.5, 0.8], [0.2, 0.8]])
        >>> mean_squared_error(input, target, sample_weight=torch.tensor([0.2, 0.8]))
        tensor(0.0650)
    """
    _mean_squared_error_param_check(multioutput)
    sum_squared_error, sum_weight = _mean_squared_error_update(
        input, target, sample_weight
    )
    return _mean_squared_error_compute(sum_squared_error, multioutput, sum_weight)


def _mean_squared_error_update(
    input: torch.Tensor,
    target: torch.Tensor,
    sample_weight: Optional[torch.Tensor],
) -> Tuple[torch.Tensor, torch.Tensor]:
    _mean_squared_error_update_input_check(input, target, sample_weight)
    return _update(input, target, sample_weight)


@torch.jit.script
def _update(
    input: torch.Tensor,
    target: torch.Tensor,
    sample_weight: Optional[torch.Tensor],
) -> Tuple[torch.Tensor, torch.Tensor]:
    squared_error = torch.square(target - input)
    if sample_weight is None:
        sum_squared_error = squared_error.sum(dim=0)
        # When input sample_weight is None, weight defaults to 1.0.
        sum_weight = torch.tensor(target.size(0), device=target.device)
    else:
        if squared_error.ndim == 2:
            sample_weight = sample_weight.unsqueeze(-1)
        sum_squared_error = (squared_error * sample_weight).sum(dim=0)
        sum_weight = sample_weight.sum(dim=0).squeeze()
    return sum_squared_error, sum_weight


@torch.jit.script
def _mean_squared_error_compute(
    sum_squared_error: torch.Tensor,
    multioutput: str,
    sum_weight: torch.Tensor,
) -> torch.Tensor:
    raw_values = sum_squared_error / sum_weight
    if multioutput == "raw_values":
        return raw_values
    else:
        return raw_values.mean()


def _mean_squared_error_update_input_check(
    input: torch.Tensor,
    target: torch.Tensor,
    sample_weight: Optional[torch.Tensor],
) -> None:
    if input.ndim >= 3 or target.ndim >= 3:
        raise ValueError(
            "The dimension `input` and `target` should be 1D or 2D, "
            f"got shapes {input.shape} and {target.shape}."
        )
    if input.size() != target.size():
        raise ValueError(
            "The `input` and `target` should have the same size, "
            f"got shapes {input.shape} and {target.shape}."
        )
    if isinstance(sample_weight, torch.Tensor) and target.size(0) != sample_weight.size(
        0
    ):
        raise ValueError(
            "The first dimension of `input`, `target` and `sample_weight` should be the same size, "
            f"got shapes {input.shape}, {target.shape} and {sample_weight.shape}."
        )


def _mean_squared_error_param_check(multioutput: str) -> None:
    if multioutput not in ("raw_values", "uniform_average"):
        raise ValueError(
            "The `multioutput` must be either `raw_values` or `uniform_average`, "
            f"got multioutput={multioutput}."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import os
from collections import defaultdict
from dataclasses import fields, is_dataclass
from typing import Any, Mapping, TypeVar

import torch
from typing_extensions import Protocol, runtime_checkable


def get_device_from_env() -> torch.device:
    """Function that gets the torch.device based on the current environment.

    This currently supports only CPU and GPU devices. If CUDA is available, this function also sets the CUDA device.

    Within a distributed context, this function relies on the ``LOCAL_RANK` environment variable
    to be made available by the program launcher for setting the appropriate device index.

    Raises:
        RuntimeError
            If ``LOCAL_RANK`` is outside the range of available GPU devices.
    """
    if torch.cuda.is_available():
        local_rank = int(os.environ.get("LOCAL_RANK", "0"))
        if local_rank >= torch.cuda.device_count():
            raise RuntimeError(
                "The local rank is larger than the number of available GPUs."
            )
        device = torch.device(f"cuda:{local_rank}")
        torch.cuda.set_device(device)
    elif torch.backends.mps.is_built() and torch.backends.mps.is_available():
        device = torch.device("mps")
    else:
        device = torch.device("cpu")
    return device


T = TypeVar("T")
TSelf = TypeVar("TSelf")


@runtime_checkable
class _CopyableData(Protocol):
    def to(self: TSelf, device: torch.device, *args: Any, **kwargs: Any) -> TSelf:
        """Copy data to the specified device"""
        ...


def _is_named_tuple(x: T) -> bool:
    return isinstance(x, tuple) and hasattr(x, "_asdict") and hasattr(x, "_fields")


def copy_data_to_device(data: T, device: torch.device, *args: Any, **kwargs: Any) -> T:
    """Function that recursively copies data to a torch.device.

    Args:
        data: The data to copy to device
        device: The device to which the data should be copied
        args: positional arguments that will be passed to the `to` call
        kwargs: keyword arguments that will be passed to the `to` call

    Returns:
        The data on the correct device
    """

    # Redundant isinstance(data, tuple) check is required here to make pyre happy
    if _is_named_tuple(data) and isinstance(data, tuple):
        return type(data)(
            **copy_data_to_device(data._asdict(), device, *args, **kwargs)
        )
    elif isinstance(data, (list, tuple)):
        return type(data)(copy_data_to_device(e, device, *args, **kwargs) for e in data)
    elif isinstance(data, defaultdict):
        return type(data)(
            data.default_factory,
            {
                k: copy_data_to_device(v, device, *args, **kwargs)
                for k, v in data.items()
            },
        )
    elif isinstance(data, Mapping):
        return type(data)(
            {
                k: copy_data_to_device(v, device, *args, **kwargs)
                for k, v in data.items()
            }
        )
    elif is_dataclass(data) and not isinstance(data, type):
        new_data_class = type(data)(
            **{
                field.name: copy_data_to_device(
                    getattr(data, field.name), device, *args, **kwargs
                )
                for field in fields(data)
                if field.init
            }
        )
        for field in fields(data):
            if not field.init:
                setattr(
                    new_data_class,
                    field.name,
                    copy_data_to_device(
                        getattr(data, field.name), device, *args, **kwargs
                    ),
                )
        return new_data_class
    elif isinstance(data, _CopyableData):
        return data.to(device, *args, **kwargs)
    return data



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, Optional, TypeVar

import torch
from torcheval.metrics.functional.classification.accuracy import (
    _accuracy_compute,
    _accuracy_param_check,
    _binary_accuracy_update,
    _multiclass_accuracy_update,
    _multilabel_accuracy_param_check,
    _multilabel_accuracy_update,
)
from torcheval.metrics.metric import Metric


TAccuracy = TypeVar("TAccuracy")
TBinaryAccuracy = TypeVar("TBinaryAccuracy")
TMultilabelAccuracy = TypeVar("TMultilabelAccuracy")


class MulticlassAccuracy(Metric[torch.Tensor]):
    """
    Compute accuracy score, which is the frequency of input matching target.
    Its functional version is ``torcheval.metrics.functional.multiclass_accuracy``.

    Args:
        average:
            - ``'micro'``[default]:
                Calculate the metrics globally.
            - ``'macro'``:
                Calculate metrics for each class separately, and return their unweighted
                mean. Classes with 0 true instances are ignored.
            - ``None``:
                Calculate the metric for each class separately, and return
                the metric for every class.
                NaN is returned if a class has no sample in ``target``.
        num_classes:
            Number of classes. Required for ``'macro'`` and ``None`` average methods.

    Example:
        >>> import torch
        >>> from torcheval.metrics import MulticlassAccuracy
        >>> metric = MulticlassAccuracy()
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5)

        >>> metric = MulticlassAccuracy(average=None, num_classes=4)
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor([1., 0., 0., 1.])

        >>> metric = MulticlassAccuracy(average="macro", num_classes=2)
        >>> input = torch.tensor([0, 0, 1, 1, 1])
        >>> target = torch.tensor([0, 0, 0, 0, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.75)

        >>> metric = MulticlassAccuracy()
        >>> input = torch.tensor([[0.9, 0.1, 0, 0], [0.1, 0.2, 0.4, 0,3], [0, 1.0, 0, 0], [0, 0, 0.2, 0.8]])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5)
    """

    def __init__(
        self: TAccuracy,
        average: Optional[str] = "micro",
        num_classes: Optional[int] = None,
    ) -> None:
        super().__init__()
        _accuracy_param_check(average, num_classes)
        self.average = average
        self.num_classes = num_classes
        if average == "micro":
            self._add_state("num_correct", torch.tensor(0.0))
            self._add_state("num_total", torch.tensor(0.0))
        else:
            # num_classes is verified to be not None when average != "micro"
            self._add_state("num_correct", torch.zeros(num_classes or 0))
            self._add_state("num_total", torch.zeros(num_classes or 0))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(self: TAccuracy, input: torch.Tensor, target: torch.Tensor) -> TAccuracy:
        """
        Update states with the ground truth labels and predictions.

        Args:
            input: Tensor of label predictions
                It could be the predicted labels, with shape of (n_sample, ).
                It could also be probabilities or logits with shape of (n_sample, n_class).
                ``torch.argmax`` will be used to convert input into predicted labels.
            target: Tensor of ground truth labels with shape of (n_sample, ).
        """
        num_correct, num_total = _multiclass_accuracy_update(
            input, target, self.average, self.num_classes
        )
        self.num_correct += num_correct
        self.num_total += num_total
        return self

    @torch.inference_mode()
    def compute(self: TAccuracy) -> torch.Tensor:
        """
        Return the accuracy score.

        NaN is returned if no calls to ``update()`` are made before ``compute()`` is called.
        """
        return _accuracy_compute(self.num_correct, self.num_total, self.average)

    @torch.inference_mode()
    def merge_state(self: TAccuracy, metrics: Iterable[TAccuracy]) -> TAccuracy:
        for metric in metrics:
            self.num_correct += metric.num_correct.to(self.device)
            self.num_total += metric.num_total.to(self.device)
        return self


class BinaryAccuracy(MulticlassAccuracy):
    """
    Compute binary accuracy score, which is the frequency of input matching target.
    Its functional version is ``torcheval.metrics.functional.binary_accuracy``.

    Args:
        threshold [default: 0.5]: Threshold for converting input into predicted labels for each sample.
        ``torch.where(input < threshold, 0, 1)`` will be applied to the ``input``.
    Example:
        >>> import torch
        >>> from torcheval.metrics import BinaryAccuracy
        >>> metric = BinaryAccuracy()
        >>> input = torch.tensor([0, 0, 1, 1])
        >>> target = torch.tensor([1, 0, 1, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.75)  # 3 / 4

        >>> metric = BinaryAccuracy(threshold=0.7)
        >>> input = torch.tensor([0, 0.2, 0.6, 0.7])
        >>> target = torch.tensor([1, 0, 1, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5)  # 2 / 4
    """

    def __init__(
        self: TBinaryAccuracy,
        threshold: float = 0.5,
    ) -> None:
        super().__init__()
        self.threshold = threshold

    @torch.inference_mode()
    def update(
        self: TBinaryAccuracy,
        input: torch.Tensor,
        target: torch.Tensor,
    ) -> TBinaryAccuracy:
        """
        Update states with the ground truth labels and predictions.

        Args:
            input: Tensor of label predictions with shape of (n_sample,).
                ``torch.where(input < threshold, 0, 1)`` will be applied to the input.
            target: Tensor of ground truth labels with shape of (n_sample,).
        """
        num_correct, num_total = _binary_accuracy_update(input, target, self.threshold)
        self.num_correct += num_correct
        self.num_total += num_total
        return self


class MultilabelAccuracy(MulticlassAccuracy):
    """
    Compute multilabel accuracy score, which is the frequency of input matching target.
    Its functional version is ``torcheval.metrics.functional.multilabel_accuracy``.

    Args:
        threshold: Threshold for converting input into predicted labels for each sample.
            ``torch.where(input < threshold, 0, 1)`` will be applied to the ``input``.
        criteria:
        - ``'exact_match'``[default]:
            The set of labels predicted for a sample must exactly match the corresponding
            set of labels in target. Also known as subset accuracy.
        - ``'hamming'``:
            Fraction of correct labels over total number of labels.

    Example:
        >>> import torch
        >>> from torcheval.metrics import MultilabelAccuracy
        >>> metric = MultilabelAccuracy()
        >>> input = torch.tensor([[0, 1], [1, 1], [0, 0], [0, 1]])
        >>> target = torch.tensor([[0, 1], [1, 0], [0, 0], [1, 1]])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5)  # 2 / 4

        >>> metric = MultilabelAccuracy(criteria="hamming")
        >>> input = torch.tensor([[0, 1], [1, 1], [0, 0], [0, 1]])
        >>> target = torch.tensor([[0, 1], [1, 0], [0, 0], [1, 1]])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.75)  # 6 / 8
    """

    def __init__(
        self: TMultilabelAccuracy,
        threshold: float = 0.5,
        criteria: str = "exact_match",
    ) -> None:
        super().__init__()
        _multilabel_accuracy_param_check(criteria)
        self.threshold = threshold
        self.criteria = criteria

    @torch.inference_mode()
    def update(
        self: TMultilabelAccuracy,
        input: torch.Tensor,
        target: torch.Tensor,
    ) -> TMultilabelAccuracy:
        """
        Update states with the ground truth labels and predictions.

        Args:
            input: Tensor of label predictions with shape of (n_sample, n_class).
                ``torch.where(input < threshold, 0, 1)`` will be applied to the input.
            target: Tensor of ground truth labels with shape of (n_sample, n_class).
        """
        num_correct, num_total = _multilabel_accuracy_update(
            input, target, self.threshold, self.criteria
        )
        self.num_correct += num_correct
        self.num_total += num_total
        return self



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, List, Optional

import torch
import torch.distributed as dist


class PGWrapper:
    """
    A wrapper around ProcessGroup that allows collectives to be issued in a
    consistent fashion regardless of the following scenarios:

        pg is None, distributed is initialized:     use WORLD as pg
        pg is None, distributed is not initialized: single process app
        pg is not None:                             use pg
    """

    def __init__(self, pg: Optional[dist.ProcessGroup]) -> None:
        if pg is None and dist.is_initialized():
            self.pg: Optional[dist.ProcessGroup] = dist.group.WORLD
        else:
            self.pg: Optional[dist.ProcessGroup] = pg

    def get_rank(self) -> int:
        if self.pg is None:
            return 0
        return dist.get_rank(group=self.pg)

    def get_world_size(self) -> int:
        if self.pg is None:
            return 1
        return dist.get_world_size(group=self.pg)

    def barrier(self) -> None:
        if self.pg is None:
            return
        dist.barrier(group=self.pg)

    # pyre-ignore[2]: obj can have arbitrary types
    def broadcast_object_list(self, obj_list: List[Any], src: int = 0) -> None:
        if self.pg is None:
            return
        dist.broadcast_object_list(obj_list, src=src, group=self.pg)

    # pyre-ignore[2]: obj can have arbitrary types
    def all_gather_object(self, obj_list: List[Any], obj: Any) -> None:
        if self.pg is None:
            obj_list[0] = obj
            return
        dist.all_gather_object(obj_list, obj, group=self.pg)


def get_process_group_backend_from_device(device: torch.device) -> str:
    """Function that gets the defaut process group backend from the device."""
    return "nccl" if device.type == "cuda" else "gloo"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

import logging
from typing import Iterable, TypeVar

import torch

from torcheval.metrics.metric import Metric

TThroughput = TypeVar("TThroughput")

_logger: logging.Logger = logging.getLogger(__name__)


class Throughput(Metric[torch.Tensor]):
    """
    Calculate the throughput value which is the number of elements processed per second.

    Note: In a distributed setting, it's recommended to use `world_size * metric.compute()`
    to get an approximation of total throughput. While using `sync_and_compute(metric)` requires
    state sync. Additionally, `sync_and_compute(metric)` will give a slightly different value compared
    to `world_size * metric.compute()`.

    Example:
        >>> import time
        >>> import torch
        >>> from torcheval.metrics import Throughput
        >>> metric = Throughput()
        >>> items_processed = 64
        >>> ts = time.monotonic()
        >>> time.sleep(2.0)  # simulate executing the program for 2 seconds
        >>> elapsed_time_sec = time.monotonic() - ts
        >>> metric.update(items_processed, elapsed_time_sec)
        >>> metric.compute()
        tensor(32.)
    """

    def __init__(self: TThroughput) -> None:
        super().__init__()
        self._add_state("num_total", torch.tensor(0.0))
        self._add_state("elapsed_time_sec", torch.tensor(0.0))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TThroughput,
        num_processed: int,
        elapsed_time_sec: float,
    ) -> TThroughput:
        """
        Update states with the values and weights.

        Args:
            num_processed: Number of items processed
            elapsed_time_sec: Total elapsed time in seconds to process ``num_processed`` items
        Raises:
            ValueError:
                If ``num_processed`` is a negative number.
                If ``elapsed_time_sec`` is a non-positive number.
        """
        if num_processed < 0:
            raise ValueError(
                f"Expected num_processed to be a non-negative number, but received {num_processed}."
            )
        if elapsed_time_sec <= 0:
            raise ValueError(
                f"Expected elapsed_time_sec to be a positive number, but received {elapsed_time_sec}."
            )

        self.elapsed_time_sec += torch.tensor(elapsed_time_sec, device=self.device)
        self.num_total += torch.tensor(num_processed, device=self.device)
        return self

    @torch.inference_mode()
    def compute(self: TThroughput) -> torch.Tensor:
        if not self.elapsed_time_sec:
            _logger.warning("No calls to update() have been made - returning 0.0")
            return torch.tensor(0.0, device=self.device)

        return self.num_total / self.elapsed_time_sec

    @torch.inference_mode()
    def merge_state(self: TThroughput, metrics: Iterable[TThroughput]) -> TThroughput:
        for metric in metrics:
            self.num_total += metric.num_total.to(self.device)
            # this assumes the metric is used within a fully-synchronous program.
            # In this scenario, the slowest process becomes the bottleneck for the
            # program's execution. As a result, we use the max, as the overall throughput
            # is gated based on the rank that takes the longest to complete.
            # TODO: should this be configurable?
            self.elapsed_time_sec = torch.max(
                self.elapsed_time_sec, metric.elapsed_time_sec.to(self.device)
            )
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, Optional, TypeVar

import torch
from torcheval.metrics.functional.classification.f1_score import (
    _f1_score_compute,
    _f1_score_param_check,
    _f1_score_update,
)
from torcheval.metrics.metric import Metric


TF1Score = TypeVar("TF1Score")


class MulticlassF1Score(Metric[torch.Tensor]):
    """
    Compute f1 score, which is defined as the harmonic mean of precision and recall.
    We convert NaN to zero when f1 score is NaN. This happens when either precision
    or recall is NaN or when both precision and recall are zero.
    Its functional version is ``torcheval.metrics.functional.multi_class_f1_score``.

    Args:
        num_classes:
            Number of classes.
        average:
            - ``'micro'``[default]:
                Calculate the metrics globally.
            - ``'macro'``:
                Calculate metrics for each class separately, and return their unweighted mean.
                Classes with 0 true and predicted instances are ignored.
            - ``'weighted'``"
                Calculate metrics for each class separately, and return their weighted sum.
                Weights are defined as the proportion of occurrences of each class in "target".
                Classes with 0 true and predicted instances are ignored.
            - ``None``:
                Calculate the metric for each class separately, and return
                the metric for every class.

    Example:
        >>> import torch
        >>> from torcheval.metrics import MulticlassF1Score
        >>> metric = MulticlassF1Score(num_classes=4)
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5000)

        >>> metric = MulticlassF1Score(average=None, num_classes=4)
        >>> input = torch.tensor([0, 2, 1, 3])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor([1., 0., 0., 1.])

        >>> metric = MulticlassF1Score(average="macro", num_classes=2)
        >>> input = torch.tensor([0, 0, 1, 1, 1])
        >>> target = torch.tensor([0, 0, 0, 0, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5833)

        >>> metric = MulticlassF1Score(num_classes=4)
        >>> input = torch.tensor([[0.9, 0.1, 0, 0], [0.1, 0.2, 0.4, 0.3], [0, 1.0, 0, 0], [0, 0, 0.2, 0.8]])
        >>> target = torch.tensor([0, 1, 2, 3])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.5)
    """

    def __init__(
        self: TF1Score,
        num_classes: Optional[int] = None,
        average: Optional[str] = "micro",
    ) -> None:
        super().__init__()
        _f1_score_param_check(num_classes, average)
        self.num_classes = num_classes
        self.average = average
        if average == "micro":
            self._add_state("num_tp", torch.tensor(0.0))
            self._add_state("num_label", torch.tensor(0.0))
            self._add_state("num_prediction", torch.tensor(0.0))
        else:
            # num_classes has been verified as a positive integer. Add this line to bypass pyre.
            assert isinstance(
                num_classes, int
            ), f"num_classes must be a integer, but got {num_classes}"

            self._add_state("num_tp", torch.zeros(num_classes))
            self._add_state("num_label", torch.zeros(num_classes))
            self._add_state("num_prediction", torch.zeros(num_classes))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(self: TF1Score, input: torch.Tensor, target: torch.Tensor) -> TF1Score:
        """
        Update states with the ground truth labels and predictions.

        Args:
            input: Tensor of label predictions.
                It could be the predicted labels, with shape of (n_sample, ).
                It could also be probabilities or logits with shape of (n_sample, n_class).
                ``torch.argmax`` will be used to convert input into predicted labels.
            target: Tensor of ground truth labels with shape of (n_sample, ).
        """
        num_tp, num_label, num_prediction = _f1_score_update(
            input, target, self.num_classes, self.average
        )
        self.num_tp += num_tp
        self.num_label += num_label
        self.num_prediction += num_prediction
        return self

    @torch.inference_mode()
    def compute(self: TF1Score) -> torch.Tensor:
        """
        Return the f1 score.

        NaN is returned if no calls to ``update()`` are made before ``compute()`` is called.
        """
        return _f1_score_compute(
            self.num_tp, self.num_label, self.num_prediction, self.average
        )

    @torch.inference_mode()
    def merge_state(self: TF1Score, metrics: Iterable[TF1Score]) -> TF1Score:
        for metric in metrics:
            self.num_tp += metric.num_tp.to(self.device)
            self.num_label += metric.num_label.to(self.device)
            self.num_prediction += metric.num_prediction.to(self.device)
        return self



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, TypeVar

import torch
from torcheval.metrics.functional.classification.auroc import (
    _auroc_compute,
    _auroc_update,
)
from torcheval.metrics.metric import Metric


TAUROC = TypeVar("TAUROC")


class BinaryAUROC(Metric[torch.Tensor]):
    """
    Compute AUROC, which is the area under the ROC Curve, for binary classification.
    Its functional version is ``torcheval.metrics.functional.binary_auroc``.

    Example:
        >>> import torch
        >>> from torcheval.metrics import BinaryAUROC
        >>> metric = BinaryAUROC()
        >>> input = torch.tensor([0.1, 0.5, 0.7, 0.8])
        >>> target = torch.tensor([1, 0, 1, 1])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.6667)
        >>> input = torch.tensor([1, 1, 1, 0])
        >>> target = torch.tensor([1, 1, 1, 0])
        >>> metric.update(input, target)
        >>> metric.compute()
        tensor(0.9167)
    """

    def __init__(
        self: TAUROC,
    ) -> None:
        super().__init__()
        self._add_state("inputs", [])
        self._add_state("targets", [])

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(
        self: TAUROC,
        input: torch.Tensor,
        target: torch.Tensor,
    ) -> TAUROC:
        """
        Update states with the ground truth labels and predictions.

        Args:
            input: Tensor of label predictions
                It should be predicted label, probabilities or logits with shape of (n_sample, ).
            target: Tensor of ground truth labels with shape of (n_samples, ).
        """
        _auroc_update(input, target)
        self.inputs.append(input)
        self.targets.append(target)
        return self

    @torch.inference_mode()
    def compute(
        self: TAUROC,
    ) -> torch.Tensor:
        return _auroc_compute(torch.cat(self.inputs), torch.cat(self.targets))

    @torch.inference_mode()
    def merge_state(self: TAUROC, metrics: Iterable[TAUROC]) -> TAUROC:
        for metric in metrics:
            if metric.inputs:
                metric_inputs = torch.cat(metric.inputs).to(self.device)
                metric_targets = torch.cat(metric.targets).to(self.device)
                self.inputs.append(metric_inputs)
                self.targets.append(metric_targets)
        return self

    @torch.inference_mode()
    def _prepare_for_merge_state(self: TAUROC) -> None:
        if self.inputs and self.targets:
            self.inputs = [torch.cat(self.inputs)]
            self.targets = [torch.cat(self.targets)]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors[16]: Undefined attribute of metric states.

from typing import Iterable, TypeVar

import torch

from torcheval.metrics.metric import Metric


TMin = TypeVar("TMin")


class Min(Metric[torch.Tensor]):
    """
    Calculate the minimum value of all elements in all the input tensors.
    Its functional version is ``torch.min(input)``.

    Example:
        >>> import torch
        >>> from torcheval.metrics import Min
        >>> metric = Min()
        >>> metric.update(torch.tensor([[1, 2], [3, 4]]))
        >>> metric.compute()
        tensor(1.)

        >>> metric.update(torch.tensor(-1)).compute()
        tensor(-1.)

        >>> metric.reset()
        >>> metric.update(torch.tensor(5)).compute()
        tensor(5.)
    """

    def __init__(self: TMin) -> None:
        super().__init__()
        self._add_state("min", torch.tensor(float("inf")))

    @torch.inference_mode()
    # pyre-ignore[14]: inconsistent override on *_:Any, **__:Any
    def update(self: TMin, input: torch.Tensor) -> TMin:
        self.min = torch.min(self.min, torch.min(input))
        return self

    @torch.inference_mode()
    def compute(self: TMin) -> torch.Tensor:
        return self.min

    @torch.inference_mode()
    def merge_state(self: TMin, metrics: Iterable[TMin]) -> TMin:
        for metric in metrics:
            self.min = torch.min(self.min, metric.min.to(self.device))
        return self
